#!/usr/bin/env python
"""
Creating The Repo Badges SVGs

Run by CI after tests and coverage, outputs SVGs into public docu dir or pushes them to
a public repo.
"""

import os
from functools import partial

import anybadge as ab
from devapp.app import app, do, run_app, system
from devapp.tools import FLG, project, dirname, exists, repl_dollar_var_with_env_val


class Flags:
    autoshort = ''

    class badges:
        """
        Requirements:
        - pypi: cfg:packagehome,e.g. in [tool.lc.app]
        - discuss: cfg:discuss
        - pipeline: cfg:repository
        - coverage: must have been run and cmd_coverage_report must deliver report
        """

        # TODO: list output not perfect at -hf:
        n = 'badges to create'
        d = [
            'documentation',
            'pypi_package',
            'discuss',
            'pipeline',
            'coverage',
            'statements',
            'code_style_ax_black',
        ]

    class badge_directory:
        """repo urls like: hg+https://...::<workdir>"""

        n = 'repo url, absolute path or relative to project root, colon sepped by link path prefix'
        d = 'public/img::img'

    class cmd_coverage_report:
        d = 'cat build/coverage_report'

    class modify_readme:
        n = 'write the badges into README.md and commit it'
        d = False

    class modify_index:
        """sometimes you want them here"""

        n = 'write the badges into docs/index.md and commit it'
        d = False

    class ignore_errors:
        n = 'exit 1 if any declard badge cannot be created'
        d = True

    class print_out:
        n = 'print markdown on exit'
        d = True


# --------------------------------------------------------------------- tools/  actions


class Repo:
    """Functions for when we are storing the badges externally in a repo"""

    def prepare_hg(repo, workdir):
        if repo.startswith('hg+'):
            repo = repo[3:]
        S.repo_url = repo
        if not exists(workdir):
            os.makedirs(dirname(workdir), exist_ok=True)
            do(system, 'hg clone "%s" "%s"' % (repo, workdir))
        else:
            do(system, 'cd "%s" && hg pull -u' % workdir)
        d = workdir + '/%s' % project.name()
        os.makedirs(d, exist_ok=True)
        return d

    def commit_and_push_if_repo():
        if not S.repo_url:
            return app.info('no repo backing of badges')
        rev = os.popen('git rev-parse --short HEAD').read().strip()
        d = project.name(), rev
        cmd = (
            "cd '%s'" % S.dest_dir,
            'hg addremove',
            "hg commit -m '%s:%s'; hg push; echo done" % d,
        )
        app.info('pushing', json=cmd)
        do(system, ' && '.join(cmd))
        ver = os.popen(cmd[0] + ' && hg identify').read().strip()
        ver = ver.split()[0]
        # https://badges.github.com/scm/hg/noauth/badges/raw-file/a07460beee0f/docutools/discuss.svg
        u = '%s/raw-file/%s/%s'
        u = u % (S.repo_url, ver, project.name())
        S.repo_badges_base_url = u
        app.info('base url for badges', url=u)


def make_directory(dir):
    pre, post = dir.split('::')
    if pre.startswith('hg+'):
        S.dest_dir = do(Repo.prepare_hg, repo=pre, workdir=post)
    else:
        dir, S.d_link_prefix = pre, post
        if not dir.startswith('/'):
            dir = os.path.join(project.root(), dir)
        S.dest_dir = dir
        os.makedirs(S.dest_dir, exist_ok=True)
    app.info('dest dir', dir=S.dest_dir)


class S:
    repo_url = None
    dest_dir = None
    markdown = []
    markdown_refs = []
    d_link_prefix = ''
    coverage_report = ''
    repo_badges_base_url = None


def read_cov_report():
    r = S.coverage_report
    if not r:
        cmd = 'cd "%s" && %s' % (project.root(), FLG.cmd_coverage_report)
        app.debug('getting coverage', cmd=cmd)
        r = S.coverage_report = os.popen(cmd).read()
    return r


def make_badge(fn, label, value, link, color, **kw):
    if not link:
        return app.warn('skipping - no link', badge=label)
    p = partial(ab.Badge, label=label, value=value, text_color='white')

    if not isinstance(color, dict):
        badge = p(default_color=color)
    else:
        badge = p(thresholds=color)
    do(badge.write_badge, file_path=fn, overwrite=True)
    il = app_cfg('homepage') + '/' + os.path.join(S.d_link_prefix, os.path.basename(fn))
    S.markdown_refs += ['[lnk_%s]: %s' % (label, link)]
    S.markdown_refs += ['[img_%s]: %s' % (label, il)]
    S.markdown += ['[![%s][img_%s]][lnk_%s]' % (label, label, label)]


class Badges:
    def documentation(fn):
        label, value, link = 'docs', app_cfg('name'), app_cfg('homepage')
        color = 'purple'
        do(make_badge, **locals())

    def pypi_package(fn):
        loc = app_cfg('packagehome')
        p = 'pypi package' if loc.startswith('https://pypi.org') else 'package'
        label, value, color = p, app_cfg('version'), '#8bd124'
        link = loc.replace('_VERSION_', app_cfg('version'))
        do(make_badge, **locals())

    def discuss(fn):
        label, value, color = ('discuss', 'skype', '#0078d4')
        link = 'https://join.skype.com/krSNYZqvEmJm'
        do(make_badge, **locals())

    def pipeline(fn):
        # we only get to here when tests finished.
        # The failing ones you'll see as repo badge, outside the markdown
        label, value, color = 'pipeline', 'passed', 'green'
        link = app_cfg('repository') + '/-/commits/master'
        do(make_badge, **locals())

    def coverage(fn, statements=False):
        s = statements
        r = do(read_cov_report)
        if not r:
            return (app.warn if FLG.ignore_errors else app.die)('no coverage report')
        r = r.rsplit('TOTAL', 1)[-1].split()[0 if s else -1]
        value = float(r[:-1]) if not s else r
        label = s['t'] if s else 'coverage'
        color = s['c'] if s else {20: 'red', 50: 'orange', 80: 'yellow', 100: 'green'}
        link = app_cfg('homepage') + '/coverage/index.html'
        do(make_badge, **locals())

    def statements(fn):
        Badges.coverage(fn, statements={'t': 'statements', 'c': 'blue'})

    def code_style_ax_black(fn):
        do(
            make_badge,
            fn,
            'code style',
            'axblack',
            'https://github.com/axiros/axblack',
            '#222222',
        )


def app_cfg(k):
    return getattr(project, k)()


modify_readme = lambda: modify_container_md(project.root() + '/README.md')
modify_index = lambda: modify_container_md(project.root() + '/docs/index.md')


md_sep = '<p attr="autogenerated by make_badges"></p>'


def modify_container_md(fn):
    with open(fn) as fd:
        md = md_orig = fd.read()
    s = md_txt_pre_code = md.split('```', 1)[0]

    if not '\n# ' in s and not s.startswith('# '):
        md = '# This package\n' + md
    l = md.split(md_sep, 2)
    if len(l) == 1:
        parts = ('\n' + md).split('\n# ', 1)
        parts[0] += '\n# '
        fh = parts[1].split('\n', 1)
        s = '''%s %s

%s

%s

        '''
        s = s % (parts[0].lstrip(), fh[0], gen_markdown(), fh[1],)
    else:
        pre, _, post = l
        s = pre + gen_markdown() + post
    if s == md_orig:
        return app.info('make_badges: no change')
    with open(fn, 'w') as fd:
        fd.write(s)
    cmd = 'git commit "%s" -m "docs: autogenerated badges" %s'
    for u, nf in (('', False), ('--author "%(USER)s"; echo true' % os.environ, True)):
        # noail since on CI we can't push anyway but dont' want to fail after a fine build.
        # docu build will have the change:
        do(system, cmd % (fn, u), no_fail=True)
    app.warn('have modified container markdown page', fn=fn)


def gen_markdown():
    if S.repo_badges_base_url:

        def replace(i, sep=']: '):
            if not i.startswith('[img_') or S.repo_badges_base_url in i:
                return i
            pre, post = i.split(sep, 1)
            post = S.repo_badges_base_url + '/' + post.rsplit('/', 1)[-1]
            return pre + sep + post

        S.markdown_refs = [replace(i) for i in S.markdown_refs]

    r = []
    _ = r.append
    _(md_sep)
    _('')
    _('&nbsp; '.join(S.markdown))
    _('')
    _('\n'.join(S.markdown_refs))
    _('')
    _(md_sep)
    return '\n'.join(r)


# ------------------------------------------------------------------------- end actions
def run():
    project.load_config()
    d = repl_dollar_var_with_env_val(FLG.badge_directory)
    make_directory(dir=d)
    # make_directory(dir=d)
    # do(make_directory, dir=FLG.badge_directory)
    for k in FLG.badges:
        try:
            do(getattr(Badges, k), fn=S.dest_dir + '/%s.svg' % k)
        except Exception as ex:
            if FLG.fail_on_error:
                raise
            app.error('Could not build', badge=k, exc=ex)
    do(Repo.commit_and_push_if_repo)
    if FLG.print_out:
        app.info(
            'printing markdown', json={'links': S.markdown, 'refs': S.markdown_refs}
        )
        print('')
        print(gen_markdown())

    do(modify_readme) if FLG.modify_readme else 0
    do(modify_index) if FLG.modify_index else 0


main = partial(run_app, run, flags=Flags)



#!/usr/bin/env python
"""
Literate Programming / Language of Babel style

Edit: We support this now not only from emacs but also from normal mkdocs markdown.
See the documentation.



Old emacs centric docs follow:

Org Mode / Mkdocs Helper Script. 

Usage: See e.g. repos/lc/docs/.../getting_started.org

Basics: 
    In you org file have this:


#+BEGIN_SRC python :exports none :results silent :session pyroot
from sys import modules as m; m.pop('lp') if 'lp' in m else 0; import lp
#lp.alias('lcm', '$d_repos/lc/build/make')
run = lp.p(lp.run, fmt='mk_cmd_out', cache='getting_started')
root = lp.p(run, root=True)
#+END_SRC

which you execute at any change of the module (reloads)

Set this:

:PROPERTIES:
:header-args:python: :session pyroot :exports results :results html
:END:

Then using is like

#+BEGIN_SRC python
root([
      {
        'cmd': 'm -l run lnr -d',
        'expect': '\n/opt/repos/lc/node-red#',
        'timeout': 15,
      },
      {
        'cmd': "ax/start -p ''",
        'expect': 'Started flows',
        'timeout': 10,
        'cmt': "-p '': no project"
      },
      ], new_session='root_nr')
#+END_SRC


You can invoke the methods also from here, via __main__
"""

import hashlib
import json
import os
import string
import subprocess as sp
import sys
import time
from functools import partial as p
from io import StringIO

env = os.environ
wait = time.sleep
now = time.time
exists = os.path.exists
user = env['USER']
d_cache = '/tmp/lp_py_cache_files_%s/' % user
env['PATH'] = 'bin:%s' % env.get('PATH', '')

cfg = {}
mk_console = '''
```%(lang)s
%(cmd)s
%(res)s
```
'''


def check_assert(ass, res):
    if ass is None:
        return
    s = str(res)
    if not isinstance(ass, (list, tuple)):
        ass = [ass]
    for a in ass:
        if not a in s:
            msg = 'Assertion failed: Expected string "%s" not found in result (\n%s)'
            raise Exception(msg % (a, s))


# seems we don't need the stupid dot anymore for the javascript xterm part (?)
# mk_cmd_out = '''

# === "Cmd"

#     ```console
#     %(cmds)s
#     ```

#     .

# === "Output"

#     %(ress)s

# '''


mk_cmd_out = '''

=== "Cmd"
    
    ```console
    %(cmds)s
    ```

=== "Output"

    %(ress)s

'''


class mk_cmd_out_fetch:
    org = '''


#+begin_tab:Command

#+BEGIN_SRC console :eval no

%(cmds)s

#+END_SRC

#+end_tab:Command


#+begin_tab:Output

%(ress)s

#+end_tab:Output

'''
    mkdocs = mk_cmd_out


xt_flat = '''
<xterm %(root)s/>

    %(ress)s

'''

xt_flat_fetch = '''
    <xterm />

         remote_content

    ![](%(fn_frm)s)
'''


# xt_flat_fetch_test = """
##+begin_tab:foo

#    <xtf %(root)s xtfrm=file:%(fn_frm)s></xtf>

##+end_tab:foo
# """

# this not yet sane:
##+begin_tab:foo

#    <xterm></xterm>

#        ansiremotecontent

#    [ ](file:./images/info_flow_adm.ansi)

##+end_tab:foo


dflt_fmt = 'mk_cmd_out'
aliases = {}
to_list = lambda s: s if isinstance(s, list) else [s]

last_result = [None]
last_ansi_file = [None]


def repl_dollar_var_with_env_val(s, die_on_fail=True):
    if not '$' in s:
        return s
    for k in env:
        if k in s:
            s = s.replace('$' + k, env[k])
    if '$' in s:
        print('Not defined in environ: $%s' % s)
    return s


def spresc(cmd):
    """subprocess run - with escape sequences escaped (for tmux)"""
    return sprun(cmd.encode('unicode-escape'))


def sprun(*a, **kw):
    """Running a command as subprocess"""
    print('--> ', a, kw)  # for the user (also in view messages)
    return p(sp.check_output, shell=True, stderr=sp.STDOUT)(*a, **kw)


class T:
    def xt_flat(resl, **kw):
        add_cmd = kw.get('add_cmd', True)
        root = '' if not kw.get('root') else 'root'
        r = []
        for cr in resl:
            cmdstr, res = get_cmd(cr), cr.get('res')
            if add_cmd and not cmdstr in res:
                res = cmdstr + '\n' + res
            r.append(res)
        ress = '\n'.join(r)
        fetch = kw.get('fetch')
        if fetch:
            fn_frm = kw['fn_frm'] = file_.write_fetchable(ress, **kw)
            t = xt_flat_fetch
            ress = t % locals()
        else:
            ress = ress.replace('\n', '\n    ')
            ress = xt_flat % locals()
        return ress

    def mk_cmd_out(res, **kw):
        root = '' if not kw.get('root') else 'root'

        def add_prompt(c, r):
            """We don't do it since console rendering requires $ or # only for syn hilite"""
            c, cmd = (c.get('cmd'), c) if isinstance(c, dict) else (c, {'cmd': c})
            if c:
                p = '#' if root else '$'
                c = '%s %s' % (p, c)
                cmt = cmd.get('cmt')
                if cmt:
                    if len(c) + len(cmt) > 80:
                        c = '%s # %s:\n%s' % (p, cmt, c)
                    else:
                        c = '%s # %s' % (c, cmt)
            return {'cmdstr': c, 'cmd': cmd, 'res': r}

        ress = T.xt_flat(res, add_cmd=kw.pop('add_cmd', False), **kw)
        res = [add_prompt(m.get('cmd'), m['res']) for m in res]
        if not any([True for m in res if m['cmdstr']]):
            return ress
        cmds = '\n'.join([r['cmdstr'] for r in res if r.get('cmdstr')])
        fetch = kw.get('fetch')
        if fetch:
            t = getattr(mk_cmd_out_fetch, kw['fetched_block_fmt'])
            r = t % locals()
        else:
            # indent one in:
            cmds = cmds.replace('\n', '\n    ')
            ress = ress.replace('\n', '\n    ')
            r = mk_cmd_out % locals()
        return r

    def mk_console(res, root=False, **kw):
        resl = res
        r = []
        lang = kw.get('lang', 'console')
        for res in resl:
            p = ''
            if lang in ['bash', 'sh']:
                p = '# ' if kw.get('root') else '$ '
            cmd = p + get_cmd(res)
            res = res['res']
            # is the command part of the res? then skip print:
            if cmd.strip() == res.strip().split('\n', 1)[0].strip():
                cmd = 'SKIP-PRINT-OUT'
            r.append(mk_console % locals())

        r = ('\n'.join(r)).splitlines()
        r = [l for l in r if not 'SKIP-PRINT-OUT' in l]
        return '\n'.join(r)


T.xtf = T.xt_flat


def get_cmd(res):
    cmd = res.get('cmd', '')
    return get_cmd(cmd) if isinstance(cmd, dict) else cmd


def alias(k, v):
    aliases[k + ' '] = v + ' '


def fmt(res, **kw):
    res = to_list(res)
    for c in res:
        c['res'] = c['res'].replace('\n```', '\n ``')
        if kw.get('hide_cmd'):
            c['cmd'] = ''

    fmt = kw.get('fmt', dflt_fmt)
    fmt = getattr(T, fmt)
    return fmt(res, **kw)


letters = string.ascii_letters + string.digits + '_'


class cache:
    """result cache, delete files or don't supply for fast execution"""

    def fn(cmd, kw):
        key = kw.get('cache')
        if not key:
            return
        s = str(cmd)
        h = hashlib.md5(s.encode('utf-8')).hexdigest()
        s = s.replace(' ', '_').replace(':', '_')
        s = ''.join([c for c in s if c in letters])
        d = d_cache + key
        os.makedirs(d, exist_ok=True)
        fn = '%s/%s_%s' % (d, s[:30], h)
        kw['fn_cache'] = fn
        return fn

    def get(cmd, kw):
        fn = cache.fn(cmd, kw)
        if not fn:
            return
        if exists(fn):
            with open(fn, 'r') as fd:
                return fd.read()

    def add(cmd, res, kw):
        fn = kw.get('fn_cache')
        if fn:
            with open(fn, 'w') as fd:
                fd.write(str(res))


def init_prompt(n):
    """run before each command"""
    # -R reset terminal state:
    sprun('tmux send-keys -R -t %s:1' % n)
    # if not mode == 'python':
    #     sprun('tmux send-keys -t %s:1 "clear" Enter' % n)
    #     while b'clear' in sprun('tmux capture-pane -ep -t %s:1' % n):
    #         wait(0.05)
    # else:
    #     pass
    #     # sprun('tmux send-keys -t %s:1 "%s" Enter' % (n, begin_cmd))

    sprun('tmux clear-history -t %s:1' % n)
    sprun("tmux send-keys -t %s:1 '' Enter" % n)


class session:
    def get(session_name, **kw):
        """
        Starts tmux if not running and delivers a srun_in_tmux function,
        parametrized for that session.
        """
        s = '\n' + os.popen('tmux ls').read()
        if not '\n%s:' % session_name in s:
            # new session:
            s = session_name
            # path is set new. bash (if executing user's shell is fish we'd be screwed)
            sprun('export SHELL=/bin/bash; export p="$PATH"; tmux new -s %s -d' % s)
            a = 'tmux send-keys -t %(session)s:1 \'export PATH="$p"; export PS1="%(prompt)s" \' Enter'
            b = {'prompt': kw.get('prompt', '$ '), 'session': s}
            for i in (1, 2):
                try:
                    sprun(a % b)
                    # the reset in init prompt needs thate time before
                    # otherwise you have the command 2 times in
                    time.sleep(0.2)
                    break
                except Exception as ex:
                    # on new systems it maybe just missing and the user / runner does not caser. Lets do it:
                    fn = env.get('HOME', '') + '/.tmux.conf'
                    if not exists(fn) and i == 1:
                        print('!! Writing %s to set base index to 1 !!' % fn)
                        r = 'set-option -g base-index 1\nset-window-option -g pane-base-index 1\n'
                        with open(fn, 'w') as fd:
                            fd.write(r)
                        continue
                    # everybody has 1 and its a mess to detect or change
                    raise Exception(
                        (
                            'tmux session start failed. Do you have tmux, '
                            'configured with base index 1? 0 is default but will NOT work!!'
                        )
                    )
            init_prompt(s)
            if kw.get('root'):
                sprun('tmux send-keys -t %s "sudo bash" Enter' % s)
                wait(0.1)

        res = p(session.srun_in_tmux, session_name=session_name)
        return res

    def srun(cmds, session_name, **kw):
        S = session.get(session_name, **kw)
        if kw.get('with_paths'):
            for e in 'PATH', 'PYTHONPATH':
                S('export %s="%s"' % (e, os.environ.get(e, '')), **kw)
        res = [{'cmd': cmd, 'res': S(cmd, **kw)} for cmd in to_list(cmds)]
        res = [i for i in res if not i.get('res') == 'silent']  # sleeps removed
        if kw.get('kill_session'):
            session.kill(session_name)
        return res

    def srun_in_tmux(cmd, session_name, expect=None, timeout=1, **kw):
        n = session_name

        # TODO: clean up
        assert_ = None
        silent = kw.get('silent')
        if isinstance(cmd, dict):
            timeout = cmd.get('timeout', timeout)
            expect = cmd.get('expect', expect)
            assert_ = cmd.get('asserts') or cmd.get('assert', assert_)
            silent = cmd.get('silent', silent)
            cmd = cmd.get('cmd')  # if not given: only produce output
        if cmd.startswith('wait '):
            time.sleep(float(cmd.split()[1]))
            return 'silent'

        sk = 'send-keys:'
        if cmd.startswith('send-keys:'):
            spresc('tmux send-keys -t %s:1 %s' % (n, cmd.split(sk, 1)[1].strip()))
            return 'silent'

        expect_echo_out_cmd = ''
        if expect is None:
            # not match on the issuing cmd
            expect_echo_out_cmd = ';echo -n ax_; echo -n done'
            cmd += expect_echo_out_cmd
            expect = 'ax_done'
        if expect is False:
            expectb = b'sollte nie vorkommen, we want timeout'
        else:
            expectb = expect.encode('utf-8')
        if cmd:
            init_prompt(n)
            spresc("tmux send-keys -t %s:1 '%s' Enter" % (n, cmd))
        t0 = now()
        wait_dt = 0.1
        while True:
            res = sprun('tmux capture-pane -epJS -1000000 -t %s:1' % n)
            if expectb in res:
                break
            if now() - t0 > timeout:
                if expect is False:
                    # wanted then:
                    break
                raise Exception(
                    'Command %s: Timeout (> %s sec) expecting "%s"'
                    % (cmd, timeout, expectb.decode('utf-8'))
                )
            wait(wait_dt)  # fast first
            wait_dt = timeout / 10.0
        res = res.decode('utf-8')
        if expect_echo_out_cmd:
            # when expect was given we include it (expect="Ready to accept Connections")
            # expect_echo_out_cmd is empty then
            res = res.split(expect, 1)[0].strip()
            res = res.replace(expect_echo_out_cmd, '')
        else:
            # the tmux window contains a lot of white space after the last output when short cmd
            res = res.strip()
        check_assert(assert_, res)

        print('----------')
        print(res)
        print('----------')
        return res if not silent else 'silent'

    def find_output_range(res, between, ls=b'\n'):
        """default: parse out last range in res between begin and end of between"""
        # todo: regex
        pre, post = res.rsplit(between[0], 1)
        r = pre.rsplit(ls, 1)[-1] + between[0]
        body, post = post.split(between[1], 1)
        r += body
        r += between[1] + post.split(ls, 1)[0]
        return r

    def kill(session_name):
        os.system('tmux kill-session -t %s' % session_name)


def pb(s):
    try:
        s = s.decode('utf-8')
    except Exception as ex:
        pass
    print(s)


def get_args(*a, **kw):
    return {'args': a, 'kw': kw}


def clear_cache():
    os.system('rm -rf "%s"' % d_cache)


def root(cmd, **kw):
    return run(cmd, root=True, **kw)


def flog(*s):
    with open('/tmp/lpcache_%s' % user, 'a') as fd:
        fd.write(str(s))
        fd.write('\n')


flog(os.getcwd())


class file_:
    def create(kw):
        fn = kw['fn']
        c = kw['content']
        if kw.get('lang') in ('js', 'javascript', 'json'):
            if isinstance(c, (dict, list, tuple)):
                c = json.dumps(c, indent=4)
        with open(fn, 'w') as fd:
            fd.write(str(c))
        os.system('chmod %s %s' % (kw.get('chmod', 660), fn))
        return file_.show(kw)

    def show(kw):
        fn = kw['fn']
        with open(fn, 'r') as fd:
            c = fd.read()
        res = {'cmd': 'cat %s' % fn, 'res': c}
        if kw.get('lang') not in ['sh', 'bash']:
            res['cmd'] = '$ ' + res['cmd']
        return res

    def write_fetchable(cont, fetch, **kw):
        """write .ansi XTF files"""
        d, fn = kw['fn_doc'].rsplit('/', 1)
        lnk = '/images/%s_%s.ansi' % (fn, fetch)
        fn = d + lnk
        if not exists(d + '/images'):
            os.makedirs(d + '/images')
        with open(fn, 'w') as fd:
            fd.write(rpl(cont, kw))
        last_ansi_file[0] = fn
        return '.' + lnk


def env_cmds(kw, when):
    cmd = kw.get(when)
    if cmd:
        if os.system(cmd):
            raise Exception('%s run failed: %s. kw: %s' % (when, cmd, str(kw)))


def run_as_python(cmd, kw):
    """
    interpret the command as python:
    no session, lang = python:
    """
    kw['fmt'] = kw.get('fmt', 'mk_console')
    h = sys.stdout
    # when a breakpoint is in a python block redirection sucks, user wants to debug:
    # TODO: write cmd to a file for better debugging
    redir = True if not 'breakpoint()' in cmd else False
    if redir:
        sys.stdout = StringIO()
    else:
        msg = '"breakpoint()" found -> not redir stdout. Result will be this message'
        print(msg)
        res = msg
    try:
        exec(cmd, {'ctx': kw})
    except Exception as ex:
        if redir:
            sys.stdout = h

    finally:
        if redir:
            try:
                res = sys.stdout.getvalue()
            except AttributeError as ex:
                res = ''  # nothing printed
            sys.stdout = h
    return {'cmd': cmd, 'res': res}


def rpl(s, kw):
    'global replacement'
    rpl = kw.get('rpl')
    if rpl:
        if not isinstance(rpl[0], (list, tuple)):
            rpl = [rpl]
        for r in rpl:
            s = s.replace(r[0], r[1])
    return s


def repl_dollar_var_with_env_vals(kw, *keys):
    for k in keys:
        v = kw.get(k)
        if v:
            kw[k] = repl_dollar_var_with_env_val(v)


def multi_line_to_list(cmd):
    if not isinstance(cmd, str):
        return cmd
    try:
        l = cmd.split('\n')
    except Exception as ex:
        print('breakpoint set')
        breakpoint()
        keep_ctx = True
    if len(l) == len([i for i in l if not i.startswith(' ')]):
        return l
    return cmd


def run(cmd, dt_cache=1, nocache=False, fn_doc=None, **kw):
    """
    dt_cache: Only cache when runtime is greater than this:
    rpl: global post run replacement
    fn_doc: required: location of source file (async flow links contain its name)
    """
    # in python sigature format assert would be forbidden, so we allow asserts=...
    assert_ = kw.get('asserts') or kw.get('assert')
    repl_dollar_var_with_env_vals(kw, 'fn', 'cwd')
    # you could set this to org:
    kw['fetched_block_fmt'] = kw.get('fetched_block_fmt', 'mkdocs')

    assert fn_doc, 'fn_doc run argument missing'
    kw['fn_doc'] = os.path.abspath(fn_doc)
    t0 = now()
    cached = cache.get(cmd, kw)
    if cached and not nocache and not env.get('disable_lp_cache'):
        return cached
    env_cmds(kw, 'cmd_prepare')

    cwd = kw.get('cwd')
    if cwd:
        here = os.getcwd()
        os.chdir(cwd)

    # mode = kw.get('xmode')
    # # in python we need tmux session (to start python first):
    # if not kw.get('session') and mode == 'python':
    #     kw['session'] = mode
    ns = kw.pop('new_session', None)
    if ns in (True, False):
        raise Exception(
            'Variable new_session must be string (the name of a session which is guaranteed a new one)'
        )
    if ns:
        session.kill(ns)
        kw['session'] = ns
    mode = kw.get('mode')
    if mode in (None, 'bash', 'ls'):
        cmd = multi_line_to_list(cmd)

    session_name = kw.pop('session', 0)
    if session_name:
        res = session.srun(cmd, session_name=session_name, **kw)
        if cwd:
            os.chdir(here)

    elif mode == 'python':
        res = run_as_python(cmd, kw)

    elif mode == 'make_file':
        kw['content'] = cmd
        kw['fmt'] = kw.get('fmt', 'mk_console')
        res = file_.create(kw)

    elif mode == 'show_file':
        kw['fmt'] = kw.get('fmt', 'mk_console')
        res = file_.show(kw)

    else:
        # os.system style:
        if not isinstance(cmd, list):
            cmd = [cmd]

        res = []
        for c in cmd:
            c1 = c['cmd'] if isinstance(c, dict) else c
            rcmd = c1
            prompt = '$'
            if kw.get('root'):
                rcmd = 'sudo %s' % rcmd
                rcmd = rcmd.replace(' && ', ' && sudo ')
                prompt = '#'
            for k, v in aliases.items():
                rcmd = rcmd.replace(k, v)

            # return sp.check_output(['sudo podman ps -a'], shell=True)
            r = sprun(rcmd)
            r = r.decode('utf-8').rstrip()
            r = ''.join([prompt, ' ', c1, '\n', r])
            # res = '%s %s\n' % (prompt, cmd) + res

            res.append({'cmd': c1, 'res': r})
    check_assert(assert_, res)
    # cwd header only for the current block:
    if cwd:
        os.chdir(here)

    if not kw.get('silent'):
        res = fmt(res, **kw)
        res = rpl(res, kw)
    else:
        res = ''

    # only cache slow stuffs:
    # if now() - t0 > dt_cache:
    cache.add(cmd, res, kw)
    last_result[0] = res
    return res


def test():
    """Better use docutools lp_client to test"""
    print(
        run(
            'cat "$HOME/bin/xdg-mime"',
            fmt='mk_cmd_out',
            fetch='program_verb2',
            fn_doc='/tmp/foo.org',
        )
    )
    return

    print(run('ls -lta', fmt='mk_cmd_out', fetch='program_verb2', fn_doc='/tmp/foo.org'))
    return
    print(run('m info', fetch='info', fn_doc='/tmp/foo.org'))
    return
    print(run('cat /usr/local/bin/m', fmt='mk_console'))
    return
    res = run({'expect': ['H: register', 'status'], 'timeout': 15,}, session='root_py')
    print(res)
    return
    run(
        [
            {
                'cmd': "m -l run lp -d -c 'echo \$CONTAINER_VER' 2>/dev/null",
                'cmt': 'version check - it is set as env var within',
                'timeout': 5,
            }
        ],
        new_session='root_py',
        fmt='mk_cmd_out',
    )
    return

    print(
        root(
            [{'cmd': 'm registry_login', 'cmt': 'foo'}, 'm -l pull lnr', 'm -l pull lp',],
            new_session='root',
            cache='foo',
        )
    )
    return

    print(
        run(
            [
                'm run -h',
                {
                    'cmd': 'm -l run lnr -Dd',
                    'expect': '\n/opt/repos/lc/node-red#',
                    'timeout': 10,
                },
                'ax/start -p false',
            ],
            new_session='root_nr',
            root=True,
        )
    )

    run('m registry_login && m -l pull lnr && m -l pull lp', root=True)
    # run(['sudo ls -lta /', 'bash', 'whoami', 'ls -lta /'], new_session='bar')


if __name__ == '__main__':
    test()



import os, json, sys, time
from devapp.tools import FLG, exists, project, write_file, read_file
from devapp.tools import define_flags, project, to_list
import inspect, shutil
from functools import partial, wraps
from .call_flow_charting import make_call_flow_chart
from theming.formatting.markdown import Mkdocs, deindent, to_min_header_level
from pprint import pformat
from inflection import humanize

from importlib import import_module
from .auto_docs import mod_doc


# ------------------------------------------------------------------------------ tools
now = time.time


class ILS:
    """
    Call Flow Logger State

    Normally populated/cleared at start and end of a pytest.
    """

    traced = set()  # all traced code objects
    max_trace = 100
    call_chain = _ = []
    counter = 0
    calls_by_frame = {}
    parents = {}
    parents_by_code = {}
    doc_msgs = []
    # sources = {}

    # wrapped = {}

    def clear():
        ILS.max_trace = 100
        ILS.call_chain.clear()
        ILS.calls_by_frame.clear()
        ILS.counter = 0
        ILS.traced.clear()
        ILS.parents.clear()
        ILS.parents_by_code.clear()
        ILS.doc_msgs.clear()


def call_flow_note(msg, **kw):
    msg = json.dumps([msg, kw], indent=2) if kw else msg
    s = {
        'input': None,
        'fn_mod': 'note',
        'output': None,
        'thread_req': thread(),
        'name': msg,
        't0': now(),
    }
    ILS.call_chain.append(s)


def unwrap_partial(f):
    while hasattr(f, 'func'):
        f = f.func
    return f


from threading import current_thread

thread = lambda: current_thread().name.replace('ummy-', '').replace('hread-', '')


class SetTrace:
    """
    sys.settrace induced callbacks are passed into this
    (when the called function is watched, i.e. added to ILS.traced)
    """

    def request(frame):
        if ILS.counter > ILS.max_trace:
            call_flow_note('Reached max traced calls count', max_trace=ILS.max_trace)
            sys.settrace(None)
            return
        ILS.counter += 1

        # need to do (and screw time) this while tracing
        co = frame.f_code
        # frm = SetTrace.get_traced_sender(frame)

        # if ILS.counter == 5:
        #     f = frame.f_back
        #     while f:
        #         print('sender', f, inp)
        #         f = f.f_back

        t0 = now()
        p = ILS.parents_by_code.get(co)
        if p:
            if len(p) == 1:
                n = p[0]
            else:
                n = '%s (%s)' % (p[-1], '.'.join(p[:-1]))
        else:
            n = co.co_name
        inp = dumps(frame.f_locals)
        # inp = ''
        s = {
            'thread_req': thread(),
            'counter': ILS.counter,
            'input': inp,
            'frame': frame,
            'parents': ILS.parents_by_code.get(co),
            'name': n,
            'line': frame.f_lineno,
            't0': now(),
            'dt': -1,
            'output': 'n.a.',
            'fn_mod': co.co_filename,
        }
        ILS.call_chain.append(s)
        ILS.calls_by_frame[frame] = s

    def response(frame, arg):
        try:
            s = ILS.calls_by_frame[frame]
        except Exception as ex:
            return
        s['dt'] = now() - s['t0']
        s['output'] = dumps(arg)
        s['thread_resp'] = thread()
        ILS.call_chain.append(s)

    def tracer(frame, event, arg, counter=0):
        """The settrace function. You can't pdb here!"""
        if not frame.f_code in ILS.traced:
            return
        if event == 'call':
            SetTrace.request(frame)
            # getting a callback on traced function exit:
            return SetTrace.tracer
        elif event == 'return':
            SetTrace.response(frame, arg)


is_parent = lambda o: inspect.isclass(o) or inspect.ismodule(o)
is_func = inspect.isfunction or inspect.ismethod


def trace_object(
    obj,
    pth=None,
    containing_mod=None,
    filters=(
        lambda k: not k.startswith('_'),
        lambda k, v: inspect.isclass(v) or is_func(v),
    ),
    blacklist=(),
):
    """recursive

    partials:
    For now we ignore them, treat them like attributes.
    The functions they wrap, when contained by a traced object will be documented,
    with all args.
    In order to document partials we would have to wrap them into new functions, set into the parent.

    filters: for keys and keys + values
    """
    if is_parent(obj):
        if not pth:
            pth = (obj.__name__,)
            ILS.parents[pth] = obj
            containing_mod = pth[0] if inspect.ismodule(obj) else obj.__module__
        for k in filter(filters[0], dir(obj)):
            v = getattr(obj, k)
            if filters[1](k, v):
                pth1 = pth + (k,)
                ILS.parents[pth1] = v
                trace_object(v, pth1, containing_mod, filters, blacklist)
    elif is_func(obj):
        spth = str(pth) + str(obj)
        if any([b for b in blacklist if b in spth]):
            # TODO: Just return in settrace and write the info into the spec of the call:
            print('blacklisted for tracing', str(obj))
            return

        if containing_mod and not obj.__module__.startswith(containing_mod):
            return
        co = obj.__code__
        ILS.parents_by_code[co] = pth
        ILS.traced.add(co)
        # print(obj)


def trace_func(traced_func, settrace=True):
    """trace a function w/o context"""
    func = unwrap_partial(traced_func)
    ILS.traced.add(func.__code__)
    # collector = collector or ILS.call_chain.append
    if settrace:
        sys.settrace(SetTrace.tracer)  # partial(tracer, collector=collector))


# we already dumps formatted, so that the js does not need to parse/stringify:
def dumps(s):
    try:
        return json.dumps(s, default=str, indent=4, sort_keys=True)
    except Exception as ex:
        # sorting sometimes not works (e.g. 2 classes as keys)
        return pformat(s)


fmts = {'mkdocs': Mkdocs}

from absl.flags._exceptions import UnparsedFlagAccessError


def autodoc_dir(mod, _d_dest):
    if _d_dest != 'auto':
        return _d_dest

    if not inspect.ismodule(mod):
        breakpoint()  # FIXME BREAKPOINT
        raise
    modn, fn = mod.__name__, mod.__file__

    r = project.root()
    if fn.startswith(r):
        # fn '/home/gk/repos/lc-python/tests/operators/test_build.py' -> tests/operators:
        d_mod = fn.rsplit(r, 1)[1][1:].rsplit('/', 1)[0]
    else:
        d_mod = modn.replace('.', '/')
    _d_dest = project.root() + '/build/autodocs/' + d_mod
    return _d_dest


flag_defaults = {}


def set_flags(flags, unset=False):
    if not unset:
        try:
            FLG.log_level  # is always imported
        except UnparsedFlagAccessError:
            from devapp.app import init_app_parse_flags

            init_app_parse_flags('pytest')
            for k in dir(FLG):
                flag_defaults[k] = getattr(FLG, k)
        M = flags
    else:
        M = flag_defaults

    return [setattr(FLG, k, v) for k, v in M.items()]


import os
from functools import partial


def pytest_plot_dest(dest):
    cur_test = lambda: os.environ['PYTEST_CURRENT_TEST']
    # if os.path.isfile(dest):
    fn_t = cur_test().split('.py::', 1)[1].replace('::', '.').replace(' (call)', '')
    dest = dest.rsplit('.', 1)[0] + '/' + fn_t

    # fn_t = cur_test().rsplit('/', 1)[-1].replace(' (call)', '')
    # '%s/%s._plot_tag_.graph_easy.src' % (dest, fn_t)
    return dest + '/_plot_tag_.graph_easy.src'


def plot_build_flow(dest):
    f = {
        'plot_mode': 'src',
        'plot_before_build': True,
        'plot_write_flow_json': 'prebuild',
        'plot_after_build': True,
        'plot_destination': partial(pytest_plot_dest, dest=dest),
    }
    return f


def add_doc_msg(msg, code=None, **kw):
    if code:
        T = fmts[kw.pop('fmt', 'mkdocs')]
        c = T.code(kw.pop('lang', 'js'), code)
        msg = (T.closed_admon if kw.pop('closed', 0) else T.admon)(msg, c, 'info')
        kw = None
    ILS.doc_msgs.append([msg, kw])


def document(
    trace,
    max_trace=100,
    fmt='mkdocs',
    blacklist=(),
    call_seq_closed=True,
    flags=None,
    dest=None,
):
    """
    Decorating call flow triggering functions (usually pytest functions) with this

    dest:
    - if not a file equal to d_dest below
    - if file: d_dest is dir named file w/o ext. e.g. /foo/bar.md -> /foo/bar/


    """

    def check_tracable(t):
        if (
            not isinstance(t, type)
            and not callable(t)
            and not getattr(type(t), '__name__') == 'module'
        ):
            raise Exception(
                'Can only trace modules, classes and callables, got: %s' % t
            )

    [check_tracable(t) for t in to_list(trace)]
    if not dest:
        # dest is empty if env var make_autodocs is not set.
        # Then we do nothing.
        # noop if env var not set, we don't want to trace at every pytest run, that
        # would distract the developer when writing tests:
        def decorator(func):
            @wraps(func)
            def noop_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return noop_wrapper

        return decorator

    ILS.max_trace = max_trace  # limit traced calls

    # this is the documentation tracing decorator:
    def use_case_deco(use_case, trace=trace):
        @wraps(use_case)
        def use_case_(*args, _dest=dest, _fmt=fmt, **kwargs):
            n_mod = use_case.__module__
            if os.path.isfile(_dest):
                fn_md_into = _dest
                _d_dest = _dest.rsplit('.', 1)[0]
                os.makedirs(_d_dest, exist_ok=True)
            else:
                raise NotImplementedError(
                    'derive autodocs dir when no mod was documented'
                )
                # _d_dest = autodoc_dir(n_mod, dest)
                # fn_md_into = (d_usecase(_d_dest, use_case) + '.md',)

            n = n_func(use_case)
            fn_chart = n
            blackl = to_list(blacklist)
            if max_trace and trace:
                [trace_object(t, blacklist=blackl) for t in to_list(trace)]
                trace_func(use_case)
            flg = {}
            if flags:
                for k, v in flags.items():
                    flg[k] = v() if callable(v) else v
                set_flags(flg)
            try:
                throw = None
                # set_flags(flags)
                value = use_case(*args, **kwargs)  # <-------- the actual original call
            except SystemExit as ex:
                throw = ex

            set_flags(flags, unset=True)
            doc_msgs = list(ILS.doc_msgs)  # will be cleared in next call:

            write.flow_chart(_d_dest, use_case, clear=True, with_chart=fn_chart)
            T = fmts[fmt]
            src = deindent(inspect.getsource(use_case)).splitlines()
            while True:
                # remove decorators:
                if src[0].startswith('@'):
                    src.pop(0)
                else:
                    break
            src = '\n'.join(src)

            # if the test is within a class we write the class as header and add the
            # funcs within:
            min_header_level, doc = 4, ''
            # n .e.g. test_01_sock_comm
            if '.' in n:
                min_header_level = 5
                section, n = n.rsplit('.', 1)
                have = written_sect_headers.setdefault(n_mod, {})
                if not section in have:
                    doc += '\n\n#### ' + section
                    have[section] = True

            doc += '\n\n'
            n_pretty = n.split('_', 1)[1] if n.startswith('test') else n
            n_pretty = humanize(n_pretty)

            # set a jump mark for the ops reference page (doc pp):
            _ = pytest_plot_dest(_dest).split('/autodocs/', 1)[1].rsplit('/', 1)[0]
            doc += '\n<span id="%s"></span>\n' % _

            ud = deindent(strip_no_ad(use_case.__doc__ or ''))
            _ = to_min_header_level
            doc += _(min_header_level, deindent('\n\n#### %s\n%s' % (n_pretty, ud)))

            n = use_case.__qualname__
            f = flg.get('plot_destination')
            if f:
                doc += add_flow_json_and_graph_easy_links(f, T)

            _ = T.closed_admon
            doc += _(n + ' source code', T.code('python', strip_no_ad(src)))

            # call flow plots only when we did trace anythning:
            if max_trace and trace:
                # m = {'fn': fn_chart}
                # svg_ids[0] += 1
                # m['id'] = svg_ids[0]
                # v = '[![](./%(fn)s.svg)](?src=%(fn)s&sequence_details=true)' % m
                # v = '![](./%(fn)s.svg)' % m  # ](?src=%(fn)s&sequence_details=true)' % m
                # v = (
                #     '''
                # <span class="std_svg" id="std_svg_%(id)s">
                # <img src="../%(fn)s.svg"></img>
                # </span>'''
                #     % m
                # )
                #![](./%(fn)s.svg)' % m  # ](?src=%(fn)s&sequence_details=true)' % m
                tr = [name(t) for t in to_list(trace)]
                tr = [t for t in tr if t]
                tr = ', '.join(tr)

                _ = call_seq_closed
                adm = T.closed_admon if _ else T.clsabl_admon
                # they are rendered and inserted at doc pre_process, i.e. later:
                fn = '%s/%s/call_flow.svg' % (fn_md_into.rsplit('.', 1)[0], fn_chart)
                fn = fn.rsplit('/autodocs/', 1)[1]
                svg_placeholder = '[svg_placeholder:%s]' % fn

                #                 svg = (
                #                     read_file('%s/%s.svg' % (os.path.dirname(fn_md_into), fn_chart))
                #                     .split('>', 1)[1]
                #                     .strip()
                #                     .split('<!--MD5', 1)[0]
                #                 )
                #                 if not svg.endswith('</svg>'):
                #                     svg += '</g></svg>'
                #                 id = '<svg id="%s" class="call_flow_chart" ' % fn_chart
                #                 svg = svg.replace('<svg ', id)
                doc += adm('Call Sequence `%s` (%s)' % (n, tr), svg_placeholder, 'info')

            # had doc_msgs been produced during running the test fuction? Then append:
            for d in doc_msgs:
                if not d[1]:
                    doc += '\n\n%s\n\n' % d[0]
                else:
                    doc += '\n%s%s' % (d[0], T.code('js', json.dumps(d[1], indent=4)))

            # append our stuffs:
            s = read_file(fn_md_into, dflt='')
            if s:
                doc = s + '\n\n' + doc
            write_file(fn_md_into, doc)

            # had the function been raising? Then throw it now:
            if throw:
                raise throw

            return value

        return use_case_

    return use_case_deco


# svg_ids = [0]


def strip_no_ad(s, sep='# /--'):
    if not sep in s:
        return s
    r, add, lines = [], True, s.splitlines()
    while lines:
        l = lines.pop(0)
        if l.strip() == sep:
            add = not add
            continue
        if not add:
            continue
        r.append(l)
    return ('\n'.join(r)).strip()


written_sect_headers = {}


def import_mod(test_mod_file):
    """
    """
    if not 'pytest' in sys.argv[0]:
        return

    if not '/' in test_mod_file:
        test_mod_file = os.getcwd() + '/' + test_mod_file
    d, fn = test_mod_file.rsplit('/', 1)
    sys.path.insert(0, d) if not d in sys.path else 0
    mod = import_module(fn.rsplit('.py', 1)[0])

    fn_md = mod_doc(mod, dest='auto')
    return fn_md


def init_mod_doc(fn_mod):
    """convenience function for test modules"""
    # cannot be done via FLG, need to parse too early:
    if not os.environ.get('make_autodocs'):
        return None, lambda: None
    fn_md = import_mod(fn_mod)
    plot = lambda: plot_build_flow(fn_md)
    return fn_md, plot


def add_flow_json_and_graph_easy_links(fn, T):
    """
    The flags had been causing operators.build to create .graph_easy files before and after build
    and also flow.json files for the before phase.

    Now add those into the markdown.
    """
    # fn like '/home/gk/repos/lc-python/build/autodocs/tests/operators/test_op_ax_socket/test01_sock_comm/_plot_tag_.graph_easy.src'
    d = os.path.dirname(fn)
    ext = '.graph_easy.src'
    # pre = fn.rsplit('/')[-1].split('_plot_tag_', 1)[0]

    def link(f, d=d):
        """Process one plot"""
        fn, l = d + '/' + f + '.flow.json', ''
        s = read_file(fn, dflt='')
        if s:
            p = '\n\n> copy & paste into Node-RED\n\n'
            l = T.closed_admon('Flow Json', p + T.code('js', s), 'info')
            # os.unlink(fn) # will be analysed by ops refs doc page
        s = f.replace(ext, '')
        dl = d.rsplit('/', 1)[-1]
        r = '\n\n![](./%s/%s.svg)\n\n' % (dl, s)
        # when s = 'test_build.py::Sharing::test_share_deep_copy.tests.post_build.svg'
        # then n = 'tests.post_build':
        if l:
            return l + r
        else:
            return T.closed_admon(s, r, 'note')

    ge = [link(f) for f in sorted(os.listdir(d)) if f.endswith(ext)]
    return '\n'.join(ge)


n_func = lambda func: func.__qualname__.replace('.<locals>', '')
d_usecase = lambda d_dest, use_case: d_dest + '/' + n_func(use_case)


def name(obj):
    qn = getattr(obj, '__name__', None)
    if qn:
        return qn
    qn = str(qn)
    return obj.replace('<', '&lt;').replace('>', '&gt;')


class write:
    def flow_chart(d_dest, use_case, clear=True, with_chart=False):
        if clear:
            sys.settrace(None)

        root_ = project.root()
        # we log all modules:
        mods = {}
        # and all func sources:
        sources = {}
        # to find input and output
        have = set()
        os.makedirs(d_usecase(d_dest, use_case), exist_ok=True)
        _ = write.arrange_frame_and_write_infos
        flow_w = [
            _(call, use_case, mods, sources, root_, d_dest, have)
            for call in ILS.call_chain
        ]
        _ = make_call_flow_chart
        fn_chart = _(flow_w, d_dest, fn=with_chart, ILS=ILS) if with_chart else 0
        # post write, clear for next use_case:
        ILS.clear() if clear else 0
        return fn_chart

    def arrange_frame_and_write_infos(
        call, use_case, mods, sources, root, d_dest, have
    ):
        """Creates
        [<callspec>, <input>, None] if input
        [<callspec>, None, <output>, None] if output
        (for the flow chart)

        and writes module and func/linenr files plus the data jsons
        """
        l = [call]
        frame = call.get('frame')
        if frame in have:
            # the output one, all other infos added already
            have.add(frame)
            l.extend([None, call.pop('output', '-')])
            return l
        have.add(frame)
        fn = call['fn_mod']
        if fn == 'note':
            return [call, 'note', None]
        d_mod = mods.get(fn)
        if not d_mod:
            d_mod = mods[fn] = write.module_filename_relative_to_root(fn, root)
            os.makedirs(d_dest, exist_ok=True)
            shutil.copyfile(fn, d_dest + '/' + d_mod[0])
        d = d_mod[0]
        call['pth'] = d_mod[1]
        call['fn_mod'] = d.rsplit('/', 1)[-1]
        func = sources.get(frame)
        if not func:
            src = inspect.getsource(frame)
            src = deindent(src)
            func_name = frame.f_code.co_name
            if 'lambda' in func_name:
                func_name = src.split(' = ', 1)[0].strip()
            fn_func = '%s.%s.func.py' % (d, call['line'])
            write_file(d_dest + '/' + fn_func, src)
            sources[frame] = func = (fn_func, func_name)
        fn_func, func_name = func
        d = d_usecase(d_dest, use_case)

        m = {
            'line': call['line'],
            'fn_func': fn_func,
            'fn_mod': d_mod[0],
            'dt': call['dt'],
            'name': func_name,
            'mod': '.'.join(d_mod[1]),
        }
        spec = [json.dumps(m), call['input'], call['output']]
        fn = call['fn'] = '%s/%s.json' % (d, call['counter'])
        write_file(fn, '\n-\n'.join(spec))
        # s = json.dumps(call, default=str)
        l.extend([call.pop('input'), None])
        return l

    def module_filename_relative_to_root(fn, root):
        d = (fn.split(root, 1)[-1]).replace('/', '__').rsplit('.py', 1)[0][2:]
        pth = d.split('__')
        return (d + '.mod.py'), pth


# def wrap_funcs_for_call_flow_logging(func_container, caller, d_into='from_pytest'):
#     """
#     Wrap all functions of container into call flow loggers, written into d_into.

#     caller: The function which invokes the call flow. Its documentation contains the call flow chart.

#     if d_into is 'from_pytest' we use the autodocs folder plus the $PYTEST_CURRENT_TEST

#     """
#     ILS.call_chain.append(('req', 0, {'name': 'intent', 'pth': ('',)}))
#     I = Inspect
#     I.decorate_with = biz_func_wrapped_for_call_flow_logging
#     ILS.mod_scan = I.recurse_into(func_container)
#     I.decorate_with = None
#     d = d_autodocs()
#     ILS.d_doc = d
#     if not exists(d):
#         os.makedirs(d, exist_ok=True)
#     write_file('%s/%s.module.py' % (d, mod.__name__), inspect.getsource(mod))


# def biz_func_wrapped_for_call_flow_logging(f, *a, _sca_=None, **kw):
#     """
#     All analysed funcs get decorated with this

#     It
#     - writes the input and result of functions into files, numbered, counting up
#     - appends itself to ILS call_chain when called
#     - and when returning

#     sca from static code analysis:
#     e.g.: {'line': 108, 'name': 'make_directory', 'pth': ('lcdoc.plugins.make_badges',)}
#     """
#     n = '%s.%s' % (_sca_['pth'][0], _sca_['line'])
#     s = ILS.sources[n]
#     if not s['written']:
#         write_file('%s/%s.func.py' % (ILS.d_doc, n), s['src'])
#         s['written'] = True
#     inp = {}
#     sig = inspect.getfullargspec(f)
#     sig_args = sig.args
#     for i in range(len(a)):
#         inp[sig_args[i]] = a[i]
#     inp.update(kw)
#     s = {'func': f, 'input': {'args': a, 'kw': kw, 'mapped': inp}}
#     s.update(_sca_)
#     ILS.counter = c = ILS.counter + 1

#     ILS.call_chain.append(('req', c, s))
#     s['t0'] = t0 = now()
#     sd = json.dumps(s, default=str)
#     try:
#         r = f(*a, **kw)
#         return r
#     except Exception as ex:
#         r = ex
#         raise
#     except SystemExit:
#         r = '[raised SystemExit]'
#         raise
#     finally:
#         dt = now() - t0
#         sdd = '%s\n%s\n%s' % (sd, json.dumps(r, default=str), dt)
#         fn = '%s/%s.json' % (ILS.d_doc, c)
#         write_file(fn, sdd)
#         s['res'] = r
#         ILS.call_chain.append(('rsp', c, s))
#     return r


# class Inspect:
#     decorate_with = None

#     def func_line_nr(func, fn, src):
#         if not callable(func):
#             return
#         for s in 'def %s(' % fn, '%s = ' % fn:
#             l = src.split(s)
#             if len(l) > 1:
#                 return len(l[0].splitlines())

#     def cls_line_nr(cls, cn, src):
#         if not isinstance(cls, type):
#             return
#         l = src.split('class %s' % cn)
#         if len(l) > 1:
#             return len(l[0].splitlines())

#     def recurse_into(obj, decorate_with=None, pth=None, **kw):
#         """static code analysis (recursive)

#         obj: module or class or func (the parent for the recursion)

#         easy when the thing is imported. there are tools which can do this with the ast.
#         but we are keen on the live sequence anyway => we'll import always anyway.
#         """
#         I = Inspect
#         pth = () if pth is None else pth
#         pobj = obj  # parent
#         g = getattr
#         while hasattr(pobj, 'func'):
#             pobj = pobj.func
#         try:
#             src = inspect.getsource(pobj)
#         except Exception as ex:
#             src = ''

#         n = getattr(pobj, '__qualname__', getattr(pobj, '__name__'))
#         if I.decorate_with and kw.get('func_no'):
#             s = '# [module %s]\n%s' % (pobj.__module__, src)
#             ILS.sources['%s.%s' % (pth[0], kw['func_no'])] = {
#                 'src': s,
#                 'written': False,
#             }

#         pth += (n,)
#         res = {'doc': pobj.__doc__ or '', 'title': n, 'source': src, pth: pth}
#         objs = [[k, g(pobj, k)] for k in dir(pobj) if not k.startswith('_')]

#         def lno(obj, n, src=src):
#             func_no = I.func_line_nr(obj, n, src) or 0
#             cls_no = I.cls_line_nr(obj, n, src) or 0
#             src = ''
#             return dict(locals())

#         objs = [lno(l[1], l[0]) for l in objs]

#         sort = lambda l, sortk: sorted(l, key=lambda k: k[sortk])
#         classes = sort([d for d in objs if d['cls_no'] > 0], sortk='cls_no')
#         funcs = sort([d for d in objs if d['func_no'] > 0], sortk='func_no')
#         r = partial(Inspect.recurse_into, pth=pth)
#         res['classes'] = [r(**c) for c in classes]
#         d = I.decorate_with
#         if d:
#             for f in funcs:
#                 func = f['obj']
#                 if func in ILS.wrapped:
#                     continue
#                 # the static code analysis results for the function we wrap:
#                 sca = {
#                     'pth': pth,
#                     'name': f['n'],
#                     'line': f['func_no'],
#                 }
#                 if not hasattr(func, 'func'):
#                     # def P(*a, _=(d, func, spec), **kw):
#                     #    return d(_[1], obj_spec=_[2], *a, **kw)
#                     P = partial(d, _sca_=sca)
#                     P = decorate(func, P)
#                 else:
#                     # partial
#                     P = partial(d, f=func, _sca_=sca)

#                 setattr(pobj, f['n'], P)
#                 ILS.wrapped[func] = (pobj, f['n'])
#         res['funcs'] = [r(**f) for f in funcs]
#         return res


# import time

# now = time.time


# def p(cls):
#     # debug helper, to print S and ILS
#     for k in dir(cls):
#         if not k.startswith('_'):
#             print('%s: %s' % (k, getattr(cls, k)))


# class flags:
#     autoshort = 'ad'
#     short_maxlen = 5

#     class pytest_autodoc:
#         """
#         Create docu pages of the current pytest function within the build dir, containing build flow charts

#         We generate the ready made markdown pages, with links to the flow charts.
#         Current format only mkdocs.

#         We only write the specs (.src) for those charts yet - since this is run while pytesting.
#         Producing charts is done later.
#         """

#         d = False


# define_flags(flags)
# from theming.formatting import markdown
# from theming.formatting.viz_sequence import Mermaid, Plant


# T = markdown.Mkdocs


# class Skip(Exception):
#     pass


# class S:
#     """state - set for each test func's call to def build_pipeline"""

#     n_mod = None
#     n_func = None
#     n_cls = None
#     mod = None
#     cls = None
#     func = None
#     fn_doc = None
#     d_root = None
#     skipped = None


# from devapp.tools import deindent


# from inflection import humanize


# details = lambda s, b: T.details % (s, b)


# def incr_header_levels(s, min_levels):
#     s = '\n' + s
#     return s.replace('\n#', '\n' + min_levels * '#')


# def add_doc(body):
#     with open(S.fn_doc, 'a') as fd:
#         fd.write(body)


# source = lambda src: details('Source code', T.py % src)

# d_autodocs_base = lambda: (S.d_root or project.root()) + '/build/autodocs'

# def set_autodoc_dir():

# def document_mod(d_dest=None):
#     """only called when mod changed"""
#     if d_dest:
#         S.fn_doc = d_dest
#     else:
#         set_autodoc_dir()

#     def find_mod(fn):
#         for k, v in sys.modules.items():
#             if (getattr(v, '__file__', None) or '').endswith(fn):
#                 return v

#     S.mod = mod = find_mod(S.n_mod)
#     # fn = lambda n: DTB() + '/' + n
#     # if not getattr(mod, 'is_auto_doc_file', 0):
#     #    raise Skip(S.n_mod)
#     S.skipped = False
#     fn = S.mod.__file__
#     S.d_root = S.d_root or project.root()
#     f = fn.split(S.d_root, 1)[1][1:].rsplit('.py', 1)[0] + '.md'
#     S.fn_doc = os.path.join(d_autodocs(), f)
#     # app.info('Generating autodoc file', fn=S.fn_doc)
#     b = mod.__doc__
#     if not b.strip():
#         b = '# %s' % mod.__name__
#     b += source(read_file(fn))
#     write_file(S.fn_doc, b, mkdir=1)


# def document_cls():
#     b = '\n\n## %s\n\n' % S.n_cls
#     S.cls = getattr(S.mod, S.n_cls)
#     b += deindent(S.cls.__doc__ or '')
#     add_doc(b)


# func_title = lambda fn: humanize(fn.split('_', 1)[1])


# def document_func():
#     l, H = (2, 'h2') if not S.n_cls else (3, 'h3')
#     h = '#' * l
#     title = func_title(S.n_func)
#     # b = '<a href="#%s">&nbsp;.</a>' % S.n_func
#     b = '\n\n' + h + ' ' + title + '\n'
#     # b = '\n\n%s %s\n###### %s\n' % ('#' * l, title, S.n_func)
#     p = S.cls if S.n_cls else S.mod
#     S.func = getattr(p, S.n_func)
#     b += incr_header_levels(deindent(S.func.__doc__ or ''), l + 1)
#     # b += add_chart()
#     # b += add_flow()
#     b += source(inspect.getsource(S.func))
#     add_doc(b)


# def parse_pytest_current_env():
#     # set for every test:
#     pyt = os.environ.get('PYTEST_CURRENT_TEST')
#     if not pyt:
#         raise Exception('This is not a pytest run - cannot generate build doc')
#     l = pyt.split('::')
#     n_mod = l[0]
#     n_cls = l[1] if len(l) > 2 else None
#     n_func = l[-1].replace('(call)', '').strip()
#     return n_mod, n_cls, n_func


# def find_what_to_document_for_new_test_func():
#     n_mod, n_cls, n_func = parse_pytest_current_env()
#     S.skipped = None  # we check that in first mod_doc and raise for all
#     funcs = []
#     # when changed then document:
#     funcs.append(document_mod) if n_mod != S.n_mod else 0
#     funcs.append(document_cls) if n_cls != S.n_cls else 0
#     funcs.append(document_func)
#     S.n_mod, S.n_cls, S.n_func = n_mod, n_cls, n_func
#     return funcs


# dirname = os.path.dirname


# def gen_func_doc():

#     """We generate the markdown docs while running pytest process
#     - we generate graph specs and write flow files for out of process analysis

#     """

#     def with_skip(f):
#         try:
#             if not S.skipped:
#                 return f()
#         except Skip:
#             print('Skipped - no mod_doc', f)
#             S.skipped = True

#     return [with_skip(f) for f in find_what_to_document_for_new_test_func()]


# def wrap_pytest_function_for_code_doc(test_func, *a, **kw):
#     """A whole new test case about to start"""
#     if ILS.wrapped:
#         raise Exception('only one call flow logging can be active at a time')
#     gen_func_doc()  # $PYTEST_CURRENT_TEST is the func arg
#     try:
#         return test_func(*a, **kw)
#     finally:
#         # do we have to gen a callgraph, i.e. did the test func call wrapped mods?
#         if ILS.wrapped:
#             ILS.make_call_flow_chart()
#             # unwraps all funcs again - will it work 100%? For sure not, we can't
#             # control time of imports... Safe is to dedicate a process for call flow logging
#             # Lets see how far we get...
#             ILS.clear()


# def gen_mod_doc(g):
#     """decorates all test funcs with gen_func_doc

#     Usage:
#     gen_mod_doc(globals()) at the end of the test module

#     """
#     if not 'pytest' in sys.argv[0]:
#         return

#     def do(n, g=g):
#         f = g[n]
#         if callable(f):
#             g[n] = decorate(f, wrap_pytest_function_for_code_doc)

#     [do(f) for f in g if f.startswith('test')]



#!/usr/bin/env python
"""
Starting Dev Server with Dir Change Watcher and Org Publisher

**Alternative** to mkdocs serve --livereload

Requires entr, tmux and npm install live-server
You have more finegrained control on used resources plus the org exporter

It will all be started in a tmux window.
"""
from plugins import _tools as tools
from devapp.app import app
import subprocess as sp
import os, sys, json, time

now = time.time

# ------------------------------------------------------------------------------ config

browsers = {'ff': 'firefox', 'gc': 'google-chrome', 'b': 'brave-browser'}


class Flags(tools.DfltFlags):
    autoshort = ''
    env_export_edit_uri = tools.env_export_edit_uri

    if 1:

        class tmux_cat_session_log:
            n = 'Display the session log'
            d = False

        class tmux_session_name:
            """We are using tmux for multicommand sessions."""

            n = 'Name of tmux session'
            d = 'docset_dev_%(dir_2_levels)s'

        class tmux_enter:
            n = 'Only attach to running session instead of creating a new one, killing the old'
            d = False

        class tmux_attach:
            """2 enables restarts of all services via a single C-b-d"""

            n = 'Attach to session after build services are up. 0: do not attach. 1: attach once. 2: re-attach after exit'
            d = 1

        class tmux_notifier:
            n = 'How to get informed about whats happening in tmux (we also write a log file)'
            d = ''

        class publish_all:
            """
            May take a while, dependent on your lob code blocks.
            """

            n = 'Publishes all org files, one by one.'
            n += 'Note: This will disable the live web server, the md rebuilding, and will set publish_disable_lob_cache to 1. Set it to -1 if you still want to render from cached results.'
            d = False

        class publish_all_excludes:
            n = 'Paths to exclude from delete-republish cycles (comma sep.ed list)'
            d = '/add/,'

        class publish_all_on_err_exit_tmux:
            n = 'By default we do not kill tmux when there is an error'
            d = False

        class publish_all_report_output_strings:
            n = 'Lowercased list of strings we should report when publishing'
            d = 'traceback, wrong, error'

        class publish_disable_lob_cache:
            n = 'Never check cached lob.py results, i.e. run all lob code'
            d = 0

    if 1:

        class server_disable:
            n = 'Do not start server'
            d = False

        class server_browser:
            n = 'Browser to use. Understood also %s' % browsers
            d = 'firefox'

        class server_no_browser:
            n = 'Add --no-browser to the server_cmd'
            d = False

        class server_wait:
            n = 'Wait ms after build'
            d = 500

        class server_cmd:
            n = 'Which server command to use. Run in dir_docset.'
            d = 'while true; do builtin cd site && live-server --browser=%(server_browser)s --wait=%(server_wait)s . ; sleep 1; done'

    if 1:

        class file_watcher_disable:
            n = 'Disable the file watcher'
            d = False

        class file_watcher_follow_symlinks:
            n = 'Follow also repos symlinks when searching for monitored files'
            d = True

        class file_watcher_watch_file_list:
            n = 'Supply a manually built watch file list'

        class file_watcher_org_publish_on_change:
            n = 'Do monitor org and publish on change'
            d = True

        class file_watcher_md_rebbuild_on_change:
            n = 'Do monitor md and rebuild on change'
            d = True

        class file_watcher_mkdoc_build_flags:
            n = 'Flags for mkdocs build'
            d = '--dirty'

        class file_watcher_fg:
            n = 'Runs only(!) the file_watcher in foreground'
            d = False


main = lambda: tools.main(run, Flags)
FLG = tools.FLG
D = tools.D
do = tools.do
system = tools.system


def build_dirs_dict():
    # for the session name:
    m = {'dir': D()}
    parent = D()
    dn, i = '', 0
    while parent:
        i += 1
        parent, d = parent.rsplit('/', 1)
        dn = d + '/' + dn
        m['dir_%s_levels' % i] = dn[:-1]
    return m


# ----------------------------------------------------------------------------- actions
tmux = tools.tmux
exists = tools.exists


class S:
    tmux = None
    stats = {'timings': {}, 'warnings': {}}


class file_watcher:
    tmemcs = None  # emacs window
    d_work = lambda: '/tmp/docs_work%s' % D()
    fn_session_log = lambda: file_watcher.d_work() + '/session.log'
    fn_stats = lambda: file_watcher.d_work() + '/stats.json'
    fn_success = lambda: file_watcher.d_work() + '/published_all_flag'

    def stats_out(stats=None):
        """When called from the tmux spawner it does not have them, but tmux _fg wrote"""
        s = stats or S.stats
        if not s:
            fn = file_watcher.fn_stats()
            if not exists(fn):
                return app.warn('no stats written yet')
            app.debug('stats from file', fn=fn)
            s = tools.read_file(json.loads(fn))
        app.info('stats', json=s)

    def start_in_tmux():
        cmd = sys.argv
        cmd.append('--file_watcher_fg')
        do(S.tmux.new_window, 'file_mon', cmd=' '.join(cmd))

    @classmethod
    def clear_work_dir(fw):
        d_work = fw.d_work()
        app.info('creating work dir', dir=d_work)
        os.makedirs(d_work, exist_ok=True)

        def rm(f):
            ff = os.path.join(d_work, f)
            os.unlink(ff)
            app.info('unlinked', fn=f)

        [rm(f) for f in os.listdir(d_work)]

    @classmethod
    def start_publish_all_fg(fw):
        f = FLG.publish_all_report_output_strings
        FLG.publish_all_report_output_strings = [
            i.strip() for i in f.split(',') if i.strip()
        ]

        def all_mds_in_mkdocs():
            # cannot import the yml due to plugins, so:
            with open('mkdocs.yml') as fd:
                mkd = fd.read().split('\nnav:', 1)[1].splitlines()
            mds = []
            d = D() + '/src/markdown/'
            while mkd:
                line = mkd.pop(0)
                lline = line.strip()
                if lline.startswith('#'):
                    continue
                if lline and not lline.startswith('#'):
                    if not line.startswith(' '):
                        break  # done

                if lline.endswith('.md'):
                    mds.append(d + lline.rsplit(':', 1)[1].strip())
            return mds

        def all_existing_orgs_in_mkdocs():
            mds = all_mds_in_mkdocs()
            orgs = [m.rsplit('.md', 1)[0] + '.org' for m in mds]
            return [o for o in orgs if exists(o)]

        t0 = now()
        excls = {i.strip() for i in FLG.publish_all_excludes.split(',') if i.strip()}
        orgs_all = all_existing_orgs_in_mkdocs()
        orgs = [o for o in orgs_all if not any([f for f in excls if f in o])]
        app.info('will publish', json=orgs, count=len(orgs), total=len(orgs_all))
        for fn in orgs:
            fw.publish_org(fn)
        s = dict(S.stats)
        pt = sorted(s['timings'].items(), key=lambda k: k[1])
        s['timings'] = {
            'total': now() - t0,
            'pages_total': sum([i[1] for i in pt]),
            'pages': pt,
        }
        with open(fw.fn_stats(), 'w') as fd:
            fd.write(json.dumps(s))
        fw.stats_out(s)
        tools.write_file(fw.fn_success(), str(now()))

    @classmethod
    def start_fg(fw):
        """We are in tmux now"""
        fw.clear_work_dir()
        if FLG.publish_all:
            return fw.start_publish_all_fg()

        app.info('watching for changes')
        do_org = FLG.file_watcher_org_publish_on_change
        do_md = FLG.file_watcher_md_rebbuild_on_change
        w = ''
        if do_org:
            w += ' -e org'
        if do_md:
            w += ' -e md'
        if not w:
            app.warn('Both watch modes disabled - not watching files')
            return
        f = '--follow' if FLG.file_watcher_follow_symlinks else ''
        get_files_cmd = "fd %s . 'src/markdown/' %s | " % (f, w)
        cmd = ['writeout () { echo "$0" ; } && ']  # % fw.fn_entr_comm_fifo()]
        cmd += ['export -f writeout; ']
        # cmd += ['while true; do ']
        cmd += [get_files_cmd]
        # -d: will exit at new file added:
        cmd += ['entr -dps writeout; sleep 0.1; ']
        # cmd += ['done ']
        cmd = ' '.join(cmd)
        t0 = now()
        while True:
            app.info('starting file monitor', cmd=cmd)
            p = sp.Popen(cmd, shell=True, stdout=sp.PIPE)
            try:
                while True:
                    try:
                        line = p.stdout.readline().decode('utf-8').strip()
                        if not line:
                            app.info('entr restarts')
                            break
                    except KeyboardInterrupt:
                        app.warn('Keyboard Interrupt')

                        testfn = '/home/gk/docs/src/markdown/add/doom.d/docutools/docs/mkdocs/index.org'
                        fw.publish_org(testfn)
                        return
                    app.info('Changed file', fn=line)
                    if now() - t0 < 1:
                        # the build took a while, when there had been other changes
                        # we get them in a flush which we ignore:
                        app.info('Ignoring file', fn=line)
                        continue

                    if line.endswith('.org'):
                        # we le the md build but the rest we ignore for a second (see a few lines above)
                        line = fw.publish_org(line)
                    if line.endswith('.md'):
                        fw.rebuild_mkdocs(line)
                    else:
                        app.die('Unsupported monitor file')
                    t0 = now()
            finally:
                p.kill()

    @classmethod
    def publish_org(fw, fn):
        is_first = False if S.stats.get('cur') else True
        S.stats['cur'] = fn  # when something failed - where is why
        tmx = fw.tmemcs
        if not tmx:
            fw.tmemcs = tmx = tmux(session_name(), window_name='emacs_publisher')

        def send_lisp(stmt, wait_for=None, timeout=1, sleep=None, warnings_into=None):
            tmx.send_keys(' \\;', enter=None)
            tmx.wait_for('Eval:')
            tmx.send_keys(stmt)
            if not wait_for:
                return
            r = None
            if warnings_into is not None:
                r = {
                    'strings': FLG.publish_all_report_output_strings,
                    'into': warnings_into,
                }
            tmx.wait_for(wait_for, timeout=timeout, sleep=sleep, monitor=r)

        def init_setup_emacs(tmx):
            m = '(global-font-lock-mode 0)'  # syn hl off
            do(send_lisp, m)
            m = "(add-to-list '+format-on-save-enabled-modes 'markdown-mode t)"
            do(send_lisp, m, wait_for='(not ', timeout=10)
            tmx.send_keys(' :', enter=None)
            tmx.send_keys('view-echo-area-messages')

        def init_setup_org_roam(tmx, fn_db='/src/markdown/org-roam.db'):
            have_roam_db = lambda: exists(D() + fn_db)
            if have_roam_db():
                app.debug('already present', fn_db=fn_db)
                return
            app.info('building first time', fn_db=fn_db)
            do(send_lisp, '(org-roam-db-build-cache)', timeout=60)
            tools.wait_for('org-roam.db', lambda: have_roam_db(), timeout=60, sleep=0.5)

        if is_first:
            t0 = now()
            do(init_setup_emacs, tmx)
            do(init_setup_org_roam, tmx)
            app.info('emacs initted', dt=now() - t0)
        warnings_into = {}
        app.debug('publishing', fn=fn)
        bn = os.path.basename
        fn_md = fn.rsplit('.org', 1)[0] + '.md'
        md_st = lambda: os.stat(fn_md)[7] if exists(fn_md) else 0
        lsp = lambda l, w, t, s, r=warnings_into: do(
            send_lisp, l, wait_for=w, timeout=t, sleep=s, warnings_into=r
        )
        lsp('(org-link-open-from-string \\"[[%s]]\\")' % fn, bn(fn), 10, 0.1)
        t0 = now()
        lsp('(org-gfm-export-to-markdown)', bn(fn_md), 60, [0.1] * 10 + [0.5])
        dt = S.stats['timings'][fn] = now() - t0
        if warnings_into:
            S.stats['warnings'][fn] = {k: list(v) for k, v in warnings_into.items()}
        app.info('published', fn=fn, took='%.2f' % dt)
        return fn_md

    def rebuild_mkdocs(line):
        cmd = 'mkdocs build --dirty'
        app.info('Triggering mkdocs rebuild', cmd=cmd, changed_file=line)
        do(system, cmd)


def start_publishing_emacs():
    app.info('Starting emacs for publishing actions')
    cmd = 'cd "%s/src/markdown" && emacs -nw index.org' % D()
    do(S.tmux.new_window, 'emacs_publisher', cmd=cmd)


def start_live_server():
    app.info('Starting live server')
    FLG.server_browser = browsers.get(FLG.server_browser, FLG.server_browser)
    if FLG.server_no_browser:
        FLG.server_browser += ' --no-browser '
    cmd = FLG.server_cmd % tools.FlagsDict(FLG)
    do(S.tmux.new_window, 'live_server', cmd)
    # when site was deleted it will keep trying. No problem
    # tools.tmux.wait_for('Serving', timeout=2, window='live_server')


session_name = lambda: FLG.tmux_session_name % build_dirs_dict()


def run():
    tools.chdir_to_docs_root()
    FLG.tmux_session_name = session_name()
    do = tools.tdo
    if FLG.file_watcher_fg:
        return file_watcher.start_fg()
    if FLG.tmux_enter:
        return do(tools.tmux.enter_session)
    if FLG.tmux_cat_session_log:
        app.info('tmux session log', fnlog=file_watcher.fn_session_log())
        print(tools.read_file(file_watcher.fn_session_log(), dflt='n.a.'))
        return
    app.warn('Starting Dev Tmux Session'.upper(), dir=D())

    if FLG.publish_disable_lob_cache > 0:
        os.environ['disable_lob_cache'] = 'true'
    else:
        os.environ['disable_lob_cache'] = ''

    do(tools.export_mk_edit_base)
    S.tmux = do(tmux.new)

    if not FLG.file_watcher_disable:
        if FLG.publish_all or FLG.file_watcher_org_publish_on_change:
            do(start_publishing_emacs)
        do(file_watcher.start_in_tmux)

    if not FLG.server_disable:
        do(start_live_server)

    app.info('All services started', tmux_session_name=FLG.tmux_session_name)
    # tmux.send_keys('alias r=\\"cd $(pwd)\\"')
    # tmux.send_keys('cd src/markdown')

    if not FLG.tmux_attach:
        app.info('Not attaching to tmux')

    elif FLG.tmux_attach == 2:
        app.info('Attaching in a loop')
        # give the user time to say ctrl-c:
        while True:
            S.tmux.call('attach')
            system(' '.join(sys.argv))
    else:
        app.info('Attaching once')
        S.tmux.call('attach')
        if FLG.publish_all:
            fw = file_watcher
            fw.stats_out()
            if not exists(file_watcher.fn_success()):
                app.die('Missing success flag')
    app.info('tmux output in', fnlog=file_watcher.fn_session_log())



from devapp.tools import exists, write_file, read_file, abspath, os, project, os
from devapp.app import app, system, do

environ = os.environ


class emacs:
    def rdp(rsc):
        root = project.root()
        DOOMDIR = root + '/cfg/doom'
        prof_name = project.root().rsplit('/', 1)[1].strip()
        # d = abspath(rsc.path + '/../')
        dde = root + '/build/doom_dot_emacs'
        return root, DOOMDIR, prof_name, dde

    def emacs(rsc, **kw):
        root, DOOMDIR, prof_name, dde = emacs.rdp(rsc)
        return {
            'cmd_pre': 'export DOOMDIR="%s"; ' % DOOMDIR,
            'cmd': 'emacs --with-profile "%s"' % prof_name,
        }

    def doom(rsc, **kw):
        root, DOOMDIR, prof_name, dde = emacs.rdp(rsc)
        return {
            'cmd_pre': 'export DOOMDIR="%s"; ' % DOOMDIR,
            'cmd': ':' + dde + '/bin/doom',
        }

    def doom_install(rsc, install=False, verify=False, api=None, **kw):
        helpme = 'https://github.com/hlissner/doom-emacs/blob/develop/docs/getting_started.org#install-doom-alongside-other-configs-with-chemacs'

        root, DOOMDIR, prof_name, dde = emacs.rdp(rsc)
        if verify:
            fn = DOOMDIR + '/init.el'
            if (
                exists(fn)
                and exists(root + '/bin/doom')
                and exists(root + '/bin/emacs')
                and exists(dde)
            ):
                app.info('doom installed', exists=fn)
                return True
            return False
        if not install:
            return
        app.info('Installing doom...')
        H = environ.get('HOME')
        fnprofs = H + '/.emacs-profiles.el'

        if not exists(fnprofs):
            if exists(H + '/.emacs'):
                app.die(
                    'you have .emacs but not .emacs-profiles.el in your home - confusing me',
                    hint='install chemacs properly',
                    see=helpme,
                )
            app.info('downloading chemacs for multiprofile mgmt')
            c = 'wget -O ~/.emacs https://raw.githubusercontent.com/plexus/chemacs/master/.emacs'
            do(system, c)
            ed = H + '/.emacs.d'
            if exists(ed):
                edn = H + '/.emacs.d.default'
                if not os.listdir(ed):
                    # emacs autocreated - in the way for chemacs
                    do(system, 'rm -f "%s"' % ed)
                app.info(
                    'moving existing emacs to chemacs default profile',
                    frm_dir=ed,
                    to_dir=edn,
                )
                do(system, 'mv "%s" "%s"' % (ed, edn))
                P = '( ("default"   . ((user-emacs-directory . "~/.emacs.default"))))'
                write_file(fnprofs, P)

        p = read_file(fnprofs)
        if '"%s"' % prof_name in p:
            msg = 'Profile name already configured in chemacs - leaving'
            app.warn(msg, name=prof_name, found_in=fnprofs)
        else:
            p = p.rsplit(')', 1)[0]
            _ = '''\n  ("%s"   . ((user-emacs-directory . "%s"))) )'''
            p += _ % (prof_name, dde)
            write_file(fnprofs, p)
        app.info('profiles content', content=p)

        cmd = 'git clone --depth 1 https://github.com/hlissner/doom-emacs "%s"' % dde

        if not exists(dde):
            do(system, cmd)
        os.chdir(dde)
        os.makedirs(DOOMDIR, exist_ok=True)
        # this is the config - ready to modify and doom will be recreatable alone from that:
        # via doom sync
        cmd = 'export DOOMDIR="%s"; bin/doom -y install' % DOOMDIR
        do(system, cmd)


class rsc:
    class emacs:
        n = 'Doom Emacs - set up 100% self contained within conda env - except one profile switcher in HOME.'
        n += 'E.g. for org file based documentation generation.'
        d = True
        cmd = emacs.emacs
        pkg = 'emacs'
        provides = emacs.doom
        post_inst = emacs.doom_install

    class graph_easy:
        n = 'Flow plotting library for terminal and svg'
        d = False
        cmd = 'graph-easy'
        pkg = 'perl(Graph::Easy)'
        conda_inst = ' && '.join(
            [
                'conda install -y -c bioconda perl-app-cpanminus ',
                'env PERL5LIB="" PERL_LOCAL_LIB_ROOT="" PERL_MM_OPT="" PERL_MB_OPT="" cpanm Graph::Easy::As_svg',
            ]
        )



import os, sys, socket, time
import inspect
from theming.formatting import markdown as MD
from devapp.tools import write_file, read_file, exists

T = MD.Mkdocs
h = lambda level, s: '#' * level + ' ' + s
source = inspect.getsource

auto_gen_cmt = '''
<!-- AUTOMATICALLY GENERATED FILE - DO NOT DIRECTLY EDIT!

Direct edits will be gone after next CI build.
By: %s@%s (%s)
Command Line (see duties.py):

    %s
-->
'''


def mark_auto_created(fn):
    from devapp.app import app

    if exists(fn):
        s = read_file(fn)
    else:
        s = fn
    h = auto_gen_cmt % (
        os.environ.get('USER'),
        socket.gethostname(),
        time.ctime(),
        ' '.join(sys.argv).replace(' -', ' \\\n     -'),
    )
    s = h.lstrip() + '\n' + s
    if exists(fn):
        write_file(fn, s)
    app.info('Marked content as autocreated')
    return s


def recurse_into(obj, md, hir=1):
    """obj a module or class"""
    code = source(obj)
    docu_obj(obj, md, hir, code)
    if hir == 1:
        md.append('## Use Cases')
        hir += 1
    # we don't recurse anymore into the module makes no sense too fragmeneted, reading the whole source is better
    # return
    # cs = classes_by_line_nr(obj, code)
    # if cs:
    #     md.append(MD.header('Classes', hir + 1))
    #     for _, c in sorted(cs.items()):
    #         recurse_into(c, md, hir + 2)

    # fs = functions_by_line_nr(obj, code)
    # if fs:
    #     md.append(MD.header('Functions', hir + 1))
    #     for _, f in sorted(fs.items()):
    #         docu_obj(f, md, hir + 2, source(f))


def name(obj):
    n = getattr(obj, '__qualname__', '') or getattr(obj, '__name__')
    return n.rsplit('.', 1)[-1]


def docu_obj(obj, md, hir, code):
    ds = obj.__doc__ or ''
    h, doc = MD.extract_docstr_head(ds)
    h = '%s' % h  # , name(obj))
    md.append(MD.header(h, hir))
    md.append(doc)
    src = T.code('python', code)
    md.append(T.closed_admon('%s source code' % name(obj), src))


def module_name(obj):
    return obj.__name__ if inspect.ismodule(obj) else obj.__module__


def in_same_module(obj1, obj2):
    """Don't document stuff outside a containing module"""
    return module_name(obj1) == module_name(obj2)


def childs(obj, typ):
    cs = [k for k in dir(obj) if not k.startswith('_')]
    cs = [getattr(obj, k) for k in cs]
    return [c for c in cs if typ(c) and in_same_module(obj, c)]


def classes_by_line_nr(obj, code):
    cs = childs(obj, inspect.isclass)
    r = {}
    for c in cs:
        line = code.split('class %s' % c.__name__, 1)
        if len(line) == 1:
            continue
        line = len(line[0].splitlines())
        r[line] = c
    return r


def functions_by_line_nr(obj, code):
    def filt(f):
        return inspect.ismethod(f) or inspect.isfunction(f)

    fs = childs(obj, filt)
    r = {}
    for f in fs:
        line = code.split('def %s' % name(f), 1)
        if len(line) == 1:
            continue
        line = len(line[0].splitlines())
        r[line] = f
    return r


def mod_doc(mod, dest='auto'):

    from lcdoc.call_flow_logging import autodoc_dir

    d = autodoc_dir(mod, dest)
    os.makedirs(d, exist_ok=True)
    fn = d + '/%s.md' % mod.__name__
    md = []
    recurse_into(mod, md)
    c = '\n'.join(md)
    write_file(fn, c)
    return fn



#!/usr/bin/env python
"""
Initialize a new mkdocs docu set

If you point to a dir with a .mkdocs.config it will patch the default yml with those.
"""
# edit from dir above this file then ide will resolve this:
from plugins import _tools as tools
from devapp.app import app
import os, sys, json
import yaml


# ------------------------------------------------------------------------------ config


class Flags(tools.DfltFlags):
    autoshort = ''

    class repos:
        n = 'Repos to include to "add" sources directory. Absolute or relative from $d_repos'
        d = ''
        t = list

    class link_templates:
        n = 'Link to the default templates instead of creating a new one. Only for doc source (this repo) development.'
        d = False

    class git_ignore_md_files:
        n = 'If you write exclusively .org this is practical for file browsing'
        d = False

    class force_build:
        n = 'Not asking for confirmations when interactive'
        d = False


main = lambda: tools.main(run, Flags)
FLG = tools.FLG
D = tools.D
do = tools.do
system = tools.system
exists = tools.exists

# ----------------------------------------------------------------------------- actions
def mk_dir_with_config():
    if exists(FLG.dir_docset + '/src'):
        app.warn('Already present', dir='src')
    cfg = tools.read_file(tools.fn_cfg(), dflt='')
    mkcfg = m = tools.read_file(D() + '/mkdocs.yml', dflt='')
    if not m:
        app.info('Found only initial mkdocs.yml')
        t = tools.fn_tmpl('mkdocs.yml')
        if FLG.link_templates:
            os.makedirs(D(), exist_ok=True)
            return tools.symlink(t, D() + '/mkdocs.yml')

        mkcfg = m = tools.read_file(tools.fn_tmpl('mkdocs.yml'))
    if cfg:
        Fubar = ': EXCLS__'
        m = m.replace(': !!', Fubar)
        m = yaml.safe_load(m)
        cfg = yaml.safe_load(cfg)
        cfg = tools.dict_merge(cfg, m)
        cfg = yaml.safe_dump(cfg)
        mkcfg = cfg.replace(Fubar, ': !!')
    tools.write_file(FLG.dir_docset + '/mkdocs.yml', mkcfg, mkdir=True)


def copy_init_src():
    frm = tools.d_installer_base + '/src'
    to = D() + '/src'
    fn = to + '/markdown/index.md'
    if exists(fn):
        return app.warn('Docu exists - skipping copying template sources', fn=fn)
    tools.tar_pipe(frm, to)


def link_repos():
    dr = D() + '/src/markdown/add'
    for r in FLG.repos:
        d = tools.full_path(r)
        if not exists(d):
            app.die('Not found', repo=d)
        s = d.rsplit('/', 1)[-1]
        # .dot files are not copied to site:
        if s.startswith('.'):
            s = s[1:]
        to_ = dr + '/' + s
        app.info('linking', repo=d, to_=to_)
        do(tools.symlink, frm=d, to=to_)


def copy_theme():
    dr = D() + '/theme'
    tools.tar_pipe(tools.d_installer_base + '/theme', dr)


def setup_git():
    dr = D()

    def add_gitignore(fn):
        if exists(fn):
            return app.warn('Already exists', fn=fn)
        ign = ['site', '*.pyc', '*.pyo', '__pycache__/', '*org-roam.db']
        if FLG.git_ignore_md_files:
            ign.append('*.md')
        with open(fn, 'w') as fd:
            fd.write('\n'.join(ign))

    if exists(dr + '/.git'):
        return app.warn('Git already present - wont create/commit', dir=dr + '/.git')
    do(system, 'git init "%s"' % dr)
    do(add_gitignore, dr + '/.gitignore')
    for fn in ['.gitignore', 'src', 'mkdocs.yml']:
        do(system, 'git add "%s"' % fn)
    do(system, 'git commit -am "Initial commit (by docuset builder)"')


# def setup_org_export():
#     return
#     t = tools.fn_tmpl('org_publish.el')
#     to = D() + '/.org_publish.el'
#     if exists(to):
#         return app.warn('skipping existing', file=to)
#     if FLG.link_templates:
#         return tools.symlink(t, to)
#     do(system, 'cp "%s" "%s"' % (t, to))


# org-protocol://roam-file?file=%2Ftmp%2Fset%2Fsrc%2Fmarkdown%2Findex.org
def setup_org_roam(check_roam_db=False):
    d_src = D() + '/src/markdown'
    fn = d_src + '/.dir-locals.el'
    if exists(fn):
        app.info('skipping existing', dir_locals=fn)
    else:
        t = tools.read_file(tools.fn_tmpl('dir-locals.el'))
        t = t.replace('_D_', D())
        tools.write_file(fn, t)


def build_docs():
    dr = D()
    do(system, '%s build' % sys.argv[0])


def final_messages():
    """
    Docset is built - ready to document!

    Next you normally do "%s dev" to start the dev services

    - Edit the mkdocs.yml for structure.
    - Check the docu (included in your build) for how to configure the browser
      links in order to open the editor on docu clicks
    """
    m = (final_messages.__doc__ % sys.argv[0]).replace('\n    ', '\n')
    m = '\n%s\n%s\n' % ('-' * 60, m)
    app.info(m)


# ----------------------------------------------------------------------------- Pipline
def run():
    m = 'Initializing new'
    if tools.exists(D() + '/mkdocs.yml'):
        m = 'Updating'

    app.warn((m + ' Doc Set').upper(), dir=D())
    if not FLG.force_build:
        if sys.stdin.isatty():
            if input('ok [y|NQ]? ').lower() != 'y':
                app.die('Unconfirmed')
        else:
            app.die('Non interactive requires force_build flag set')
    # add docs repo:
    d = tools.dir_of(tools.d_installer_base, up=1)
    assert '.git' in os.listdir(d), '.git not found in ' + d
    FLG.repos.append(d)
    do = tools.tdo
    do(mk_dir_with_config)
    FLG.dir_docset = D()
    os.chdir(D())
    do(copy_init_src)
    do(link_repos)
    do(copy_theme)
    # do(setup_org_export)
    do(setup_git)
    do(setup_org_roam)
    do(build_docs)
    do(final_messages)



#!/usr/bin/env python
from devapp.tools import (
    FLG,
    FlagsDict,
    read_file,
    write_file,
    dir_of,
    dict_merge,
    partial,
)
from devapp.app import app, run_app, system, do  # noqa
import os, sys, json, time
from dataclasses import dataclass

wait, now = time.sleep, time.time


# ------------------------------------------------------------------------------ config
exists = os.path.exists
d_start = os.getcwd()
d = d_start
while len(d) > 2:
    if exists(d + '/mkdocs.yml'):
        break
    d = d.rsplit('/', 1)[0]
d_dflt = d if len(d) > 2 else d_start + '/docs'

# used in dev and build:
class env_export_edit_uri:
    n = 'Export $mk_edit_base, used via env var in mkdocs.yml file'
    d = 'org-protocol://roam-file?file=%(dir)s/src/markdown'


class BaseFlags:
    class dir_docset:
        n = '''Directory of docuset (location of its mkdocs.yml).
            Full path or from $d_repo.
            Can contain a .mkdocs.config'''
        d = d_dflt


class DfltFlags(BaseFlags):
    class tmux_base_index:
        """
        base_index 1 means: Start windows and panes at 1, not 0.
        """

        n = 'Set to 0 if your tmux conf has NOT the usual altered base index set to 1.'
        d = 1


def main(run, Flags, log_level=10):
    return run_app(run, flags=Flags, kw_log={'level': log_level})


# ------------------------------------------------------------------------------- tools
here = dir_of(__file__)
d_installer_base = dir_of(here, up=1) + '/docutools/mkdocs'


D = lambda: full_path(FLG.dir_docset)
fn_cfg = lambda: D() + '/.mkdocs.config'
fn_mkdocs_yml = lambda: D() + '/mkdocs.yml'
fn_tmpl = lambda t: d_installer_base + '/templates/%s' % t


def chdir_to_docs_root():
    if not exists(fn_mkdocs_yml()):
        app.die('Cannot derive your docset directory')
    os.chdir(D())


def export_mk_edit_base():
    value = FLG.env_export_edit_uri % {'dir': D()}
    # value = 'org-protocol://roam-file?file=%s/src/markdown' % value
    app.info(
        'exporting $mk_edit_base for mkdocs edit links',
        used_in='mkdocs.yml',
        value=value,
    )
    os.environ['mk_edit_base'] = value


def full_path(fn):
    if os.path.exists(fn):
        fn = os.path.abspath(fn)
    if fn.startswith('/'):
        return fn
    if fn.startswith('.'):
        return os.path.abspath(fn)
    d = os.environ['d_repos']
    return d + '/' + fn


def tar_pipe(frm, to):
    if not exists(to):
        os.mkdir(to)
    do(system, '(cd "%s" && tar -cf - .) | (cd "%s" && tar -xpf -)' % (frm, to))


tdo = partial(do, titelize=True)


def symlink(frm, to, rmdir=False):
    if os.path.islink(to):
        d = os.readlink(to)
        if d != frm:
            app.warn('leaving existing but different link', at=to, points=d, should=frm)
        else:
            app.debug('link exists already', to=to, frm=frm)
        return
    if os.path.exists(to):
        if rmdir:
            app.warn('removing', target=to)
            assert len(to.split('/') > 4)
            do(system, '/bin/rm -rf "%s"' % to)
    do(system, 'ln -s "%s" "%s"' % (frm, to))


def wait_for(msg, f=None, *a, timeout=1, sleep=None, **kw):
    """wait for truthy return of function or, alternative (errcode, value) result"""
    if not isinstance(sleep, list):
        sleep = [timeout / 10.0] if sleep is None else [sleep]
    t0 = now()
    while True:
        res = f(*a, **kw)
        # print(res)
        if res:
            # in errcode, value format?
            if isinstance(res, tuple):
                if not res[0]:
                    return res[1]
            else:
                return res
        if now() - t0 > timeout:
            break
        ls = sleep.pop(0) if sleep else ls
        wait(ls)
    app.die('Timeout waiting', for_=msg, waited=now() - t0, json={'got result': res})


@dataclass
class tmux:
    """
    A tmux session or window within session
    Use get_tmux to get a new parametrized class.
    """

    session_name: str
    window_name: str = ''  # optional

    @classmethod
    def new(cls, session_name=None, first_win='docset', kill_any_old=True):
        sn = session_name or FLG.tmux_session_name
        s = partial(do, system, no_fail=True)
        if kill_any_old:
            s('tmux kill-session -t "%s"' % sn)
        s('tmux new -s "%s" -d -n docset' % sn)
        t = tmux(session_name=sn)
        app.info('crea new', **t.__dict__)
        return t

    def call(self, cmd, args='', no_fail=False):
        sn = self.session_name
        if self.window_name:
            sn += ':%s' % self.window_name
        tcmd = 'tmux ' + cmd + ' -t "%s" ' % sn + args
        do(system, tcmd)

    def new_window(self, name, cmd=None):
        self.call('new-window', '-n "%s"' % name)
        if cmd:
            self.send_keys(cmd)

    def send_keys(self, keys, enter=True):
        keys = '"%s"' % keys
        if enter:
            keys += ' Enter'
        return self.call('send-keys', keys)

    def enter_session(self):
        return self.call('att')

    def capture(self, window=None):
        i = self.session_name
        w = window or self.window_name
        if w:
            i += ':%s' % w
        res = os.popen('tmux capture-pane -p -t "%s"' % i).read()
        return res

    def wait_for(self, content, timeout=1, sleep=None, window=None, monitor=None):
        def f(window=window, monitor=monitor):
            # no_capture can be passed when not-understood timeouts happen, in order
            # to eval the last result
            cap = self.capture(window)
            if monitor:
                for s in monitor['strings']:
                    capl = cap.lower()
                    if s in capl:
                        pre, post = capl.split(s, 1)
                        line = pre.rsplit('\n', 1)[-1] + s + post.split('\n', 1)[0]
                        h = monitor['into'].setdefault(s, set())
                        if not line in h:
                            app.warn('found: "%s"' % s, line=line)
                            h.add(line)
            if content in cap:
                app.debug('found expected content', content=content)
                return 0, cap
            return 1, cap.strip()

        wait_for(content, f, timeout=timeout, sleep=sleep)



import sys, os
from lcdoc.auto_docs import mod_doc
from importlib import import_module
from devapp.tools import FLG


# def gen_mod_doc(test_mod_file):
#     """
#     """
#     if not 'pytest' in sys.argv[0]:
#         return

#     if not FLG.make_autodocs:
#         return
#     if not '/' in test_mod_file:
#         test_mod_file = os.getcwd() + '/' + test_mod_file
#     d, fn = test_mod_file.rsplit('/', 1)
#     sys.path.insert(0, d) if not d in sys.path else 0
#     mod = import_module(fn.rsplit('.py', 1)[0])

#     fn_md = mod_doc(mod, dest='auto')
#     return fn_md


# from devapp.app import app
# import os, json, sys
# from devapp.tools import (
#     FLG,
#     filter_passwords,
#     exists,
#     project,
#     write_file,
#     read_file,
#     dirname,
# )
# from devapp.tools import define_flags
# import inspect


# from theming.formatting import markdown

# T = markdown.Mkdocs


# class Skip(Exception):
#     pass


# class S:
#     """state - set for each test func's call to def build_pipeline"""

#     n_mod = None
#     n_func = None
#     n_cls = None
#     mod = None
#     cls = None
#     func = None
#     fn_doc = None
#     d_root = None


# from devapp.tools import deindent


# from inflection import humanize


# details = lambda s, b: T.details % (s, b)


# def incr_header_levels(s, min_levels):
#     s = '\n' + s
#     return s.replace('\n#', '\n' + min_levels * '#')


# def add_doc(body):
#     with open(S.fn_doc, 'a') as fd:
#         fd.write(body)


# source = lambda src: details('Source code', T.py % src)


# def do_mod():
#     def find_mod(fn):
#         for k, v in sys.modules.items():
#             if (getattr(v, '__file__', None) or '').endswith(fn):
#                 return v

#     S.mod = mod = find_mod(S.n_mod)
#     # fn = lambda n: DTB() + '/' + n
#     if not getattr(mod, 'is_auto_doc_file'):
#         raise Skip(S.n_mod)
#     fn = S.mod.__file__
#     S.d_root = S.d_root or project.root()
#     f = fn.split(S.d_root, 1)[1][1:].rsplit('.py', 1)[0] + '.md'
#     S.fn_doc = os.path.join(S.d_root + '/build/autodocs', f)
#     app.info('Generating autodoc file', fn=S.fn_doc)
#     b = mod.__doc__
#     if not b.strip():
#         b = '# %s' % mod.__name__
#     b += source(read_file(fn))
#     write_file(S.fn_doc, b, mkdir=1)


# def do_cls():
#     b = '\n\n## %s\n\n' % S.n_cls
#     S.cls = getattr(S.mod, S.n_cls)
#     b += deindent(S.cls.__doc__ or '')
#     add_doc(b)


# func_title = lambda fn: humanize(fn.split('_', 1)[1])


# def do_func():
#     l, H = (2, 'h2') if not S.n_cls else (3, 'h3')
#     h = '#' * l
#     title = func_title(S.n_func)
#     # b = '<a href="#%s">&nbsp;.</a>' % S.n_func
#     b = '\n\n' + h + ' ' + title + '\n'
#     # b = '\n\n%s %s\n###### %s\n' % ('#' * l, title, S.n_func)
#     p = S.cls if S.n_cls else S.mod
#     S.func = getattr(p, S.n_func)
#     b += incr_header_levels(deindent(S.func.__doc__ or ''), l + 1)
#     # b += add_chart()
#     # b += add_flow()
#     b += source(inspect.getsource(S.func))
#     add_doc(b)


# def register_new_test_func(pyt):
#     l = pyt.split('::')
#     funcs = []
#     mod = l[0]
#     funcs.append(do_mod) if mod != S.n_mod else 0
#     S.n_mod = mod
#     cls = None
#     if len(l) > 2:
#         cls = l[1]
#         funcs.append(do_cls) if cls != S.n_cls else 0
#     S.n_cls = cls
#     S.n_func = l[-1].replace('(call)', '').strip()
#     funcs.append(do_func)
#     return funcs


# def gen_func_doc():

#     """We generate the markdown docs while running pytest process
#     - we generate graph specs and write flow files for out of process analysis

#     """
#     pyt = os.environ.get('PYTEST_CURRENT_TEST')
#     if not pyt:
#         app.die('This is not a pytest run - cannot generate build doc')
#     breakpoint()  # FIXME BREAKPOINT
#     return [f() for f in register_new_test_func(pyt)]


# from decorator import decorate



#!/usr/bin/env python
"""
Testing lob.py snippets

"""
from plugins import _tools as tools
from devapp.app import app
import os, sys, json


# ------------------------------------------------------------------------------ config


class Flags(tools.BaseFlags):
    autoshort = ''

    class snippet:
        n = 'Snippet code or file'
        d = """run('ls -lta / --color=always', fmt='xt_flat')"""

    class sh_ansi_file:
        n = 'Output created ansi files'
        d = False


main = lambda: tools.main(runit, Flags)
FLG = tools.FLG
D = tools.D
exists = tools.exists
do = tools.do
system = tools.system
dirname = os.path.dirname
# ----------------------------------------------------------------------------- actions


def set_sys_path():
    pp = os.environ['HOME'] + '/.doom.d/bin'
    if not pp in sys.path:
        sys.path.insert(0, pp)
    app.info('added (like emacs in config.el) to sys.path', dir=pp)


class S:
    pass


import importlib

T = """
import sys, os
pp = os.environ['HOME'] + '/.doom.d/bin'
if not pp in sys.path:
    sys.path.insert(0, pp)
import lob
run= lob.p(lob.run, fn_org="%s/foo.org")
%s """


def run_snippet():
    import lob

    d_test = '/tmp/sniptests'
    s = FLG.snippet
    if os.path.exists(s):
        app.info('snippet is a file', reading=s)
        s = tools.read_file(s)
    S.code = s
    S.mod_code = T % (d_test, S.code)

    os.makedirs(d_test, exist_ok=True)
    if d_test.startswith('/tmp/'):  # fear...
        os.system('/bin/rm -rf "%s/*"' % d_test)
    else:
        raise
    sys.path.insert(0, d_test)

    fnm = d_test + '/test_snippet.py'
    app.info('writing', test_mod_file=fnm)
    tools.write_file(fnm, S.mod_code)
    app.info('importing test_mod_file', content=S.mod_code)
    mod = importlib.import_module('test_snippet')
    app.info('Result', result=mod.lob.last_result[0])
    l = mod.lob.last_ansi_file[0]
    if l:
        app.info('.ansi file was created', fn=l)
    if FLG.sh_ansi_file:
        c = tools.read_file(l)
        app.info('.ansi file content', content=c)


def runit():
    app.warn('Running lob code snippet'.upper(), dir=D(), snip=FLG.snippet)
    do(run_snippet)



#!/usr/bin/env python
"""
Passthrough to Standard Mkdocs

"""
import os, sys


def main():
    if sys.argv[0] != __file__:
        sys.argv.pop(1)
    sys.exit(os.system('mkdocs %s' % ' '.join(sys.argv[1:])))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Allows to run the full cycle for autodocs generation

"""
from plugins import _tools as tools
from devapp.app import app
from devapp import tools as datools
import os, sys, json


# ------------------------------------------------------------------------------ config


class Flags(tools.DfltFlags):
    autoshort = ''

    class clear:
        n = 'Clear the autodocs folder'
        d = False

    class flags:
        n = 'If pytest should run, supply the flags here'

    class pre_proc:
        n = 'How to run doc pre_process'

    class mkdocs:
        """Complete Example: doc pt -f "-xs tests/operators/test_build.py" -pp -gad -m"""

        n = 'Run mkdocs build'
        d = False


main = lambda: tools.main(run, Flags)
FLG = tools.FLG
D = tools.D
do = tools.do
system = tools.system

# ----------------------------------------------------------------------------- actions


def run():
    root = datools.project.root()
    if FLG.clear:
        do(system, '/bin/rm -rf "%s/build/autodocs/*"' % root)
    f = FLG.flags
    if f:
        app.warn('Setting $make_autodocs=true -> triggers call flow logging')
        os.environ['make_autodocs'] = 'true'
        e = do(system, 'pytest -p no:randomly %s' % f)
    f = FLG.pre_proc
    if f:
        do(system, 'doc pre_process %s' % f)
    f = FLG.mkdocs
    if f:
        do(system, 'cd "%s" && mkdocs build' % root)



from theming.formatting import markdown
from theming.formatting.viz_sequence import Mermaid, Plant
from devapp.tools import write_file, project
import json

vizs = {'plantuml': Plant, 'mermaid': Mermaid}


def make_call_flow_chart(flow, d_dest, fn=None, viz_name='plantuml', ILS=None):
    """
    - flow from original ILS.call_chain (list of calls, one req, one resp)
    - reformatted for [ [{<call n>}, <input>, <output>], ...] by arrange_call_and_write_infos

    call: traced frame (frame incl. meta infos as traced)

    ILS: Reference to the call_flow_logging.ILS (all the recorded state is there)
    """
    # breakpoint()  # FIXME BREAKPOINT
    fn = 'call_seq' if fn is None else fn
    r = []
    if not flow:
        print('no flow')
        return
    r_add = r.append
    viz = vizs[viz_name]
    d_root = project.root()
    # flow.insert(
    #    0, [{'counter': 0, 'name': 'intent', 'pth': ('outside',)}, 'null', 'null']
    # )
    last_thread = [0]

    def add_thread(call, have=set(), r_add=r_add, last=last_thread):
        thread = call['thread_req']
        if thread in have:
            return
        have.add(thread)
        last[0] = thread
        r_add(viz.participant(thread, thread))

    # cc is created in biz_func_wrapped_for_call_flow_logging (typ, nr, data)
    # flow.insert(0, [{'thread_req': 't'}])
    [add_thread(c[0]) for c in flow]
    # flow.pop(0)

    def args_str_and_tooltip(c, viz=viz):
        if c == 'null':
            return '-', '(None)'
        try:
            args = json.loads(c)
        except Exception as ex:
            return c.replace('\n', ','), c
        s = ['%s:%s' % (str(k), str(v)) for k, v in args.items()]
        t = '\n'.join(s)
        s = ','.join(s)
        if not s.strip():
            return 'null', 'null'
        return s, t

    def trim(s, distance, viz=viz, chars_between_adjancent_boxes=20):
        """The displayed call args in the links - should be not too long to not
        widen the chart yet more"""
        d = (abs(distance) + 1) * chars_between_adjancent_boxes
        s = s if len(s) < d else s[: d - 2] + '..'
        # s = viz.ls.join(s.splitlines())
        for k in '{', '[', '}', ']':
            s = s.replace(k, ' ')
        return s

    def href(s, tooltip, nr, typ, viz=viz, max_tooltip_len=3000):
        if not s:
            return ''
        # [[http://plantuml.com/start{Tooltip for message} some link]]
        s = s.replace('\n', ' ')  # text over arrows should be not multline
        t = tooltip.replace('\n', viz.ls)
        t = t[:max_tooltip_len]  # bigger? he should click then
        t = t.replace('<', ' ').replace('>', '')
        r = '%s.json?%s{%s} %s' % (nr, typ, t, s)
        # r = 'javascript:onmouseover(x=>alert("foo")){foo} foo'
        r = r.replace(']]', '\\]\\]')
        return '[[%s]]' % r

    frm = []  # last req sender, the stack of active functions
    # i = [thread_name(f[0]) for f in flow]
    # breakpoint()  # FIXME BREAKPOINT
    t0 = flow[0][0]['t0']
    tn = flow[-1][0]['t0'] + flow[-1][0].get('dt', 0)
    DT = tn - t0
    for rnd in range(-1, 6):
        j = 1 * round((DT / 10.0), rnd)
        if j != 0:
            break
    dt = j
    time_ = 0
    while flow:
        call, inp, output = flow.pop(0)
        t0 = call['t0']
        thread = call['thread_req']
        if call.get('fn_mod') == 'note':
            msg = call['name'].replace('\n', viz.ls)
            r_add('note over %s: %s' % (thread, msg))
            continue
        nr = call['counter']
        n = call['name']
        if inp:
            if t0 > time_:
                # r_add('d <-> t: %s' viz.req('d', 't', what=str(round(time_, rnd))))
                #% (last_thread[0], round(time_, rnd))
                r_add('[-[#888888]-> MainThread: %s' % round(time_, rnd))
                time_ += dt
            n_with_url = href(n, str(inp), nr, 'req')
            r_add('%s<-[#888888]-]: %s' % (thread, n_with_url))
            # r_add('note left of %s: %s' % (thread, n))
            # r_add('[-[#000033]-> %s: %s' % (thread, ''))
            r_add('activate %s' % thread)

            # if frm:
            #     nfrm = frm[-1]
            #     # breakpoint()  # FIXME BREAKPOINT
            #     # ff = call.get('from_frame')
            #     # if ff:
            #     # nfrm = particips_shorts_by_code[ff.f_code]
            #     # if '"return msg.reconfigure ? [msg, null] : [null, msg]"' in str(inp):
            #     #    breakpoint()  # FIXME BREAKPOINT
            #     s, t = args_str_and_tooltip(inp)
            #     s1 = trim(s, distance=(nfrm - n))
            #     inp = href(s1, t, nr, 'req')
            #     # add the arrow form the calling function to us, with the req params:
            #     r_add(viz.req(nfrm, n, what=inp))
            # add(viz.activate(n))
            # frm.append(n)
        else:
            r_add('deactivate %s' % thread)
            # continue
            # assert inp == None
            # # when a source produces like an observer.on_next loop, the function has
            # # an entry but never exits. Then only an arrow goes out of the func:
            # while frm[-1] != n:
            #     frm.pop()
            # frm.pop()
            # w = None if output == 'null' else output
            # if not frm:
            #     break
            # if w:
            #     s = t = w
            #     w = trim(w, distance=(n - frm[-1]))
            # else:
            #     w = 'null'
            # # add the arrow back to the calling function, with the response
            # r_add(viz.rsp(n, frm[-1], what=href(w, t, nr, 'resp') or ''))
            # # add(viz.deactivate(n))

    fn = '%s/%s/call_flow.%s' % (d_dest, fn, viz.ext)
    r = viz.wrap(r)  # e.g. @startuml, and concat to string
    write_file(fn, r)
    return fn

    # loc = S.d_root + '/build/autodocs/'
    # m = {
    #     'fn': fn.split(dirname(S.fn_doc), 1)[1].rsplit('.', 1)[0],
    #     'loc': len(ILS.d_doc.split(loc, 1)[1].split('/')) * '../',
    # }
    # if viz == Plant:
    #     # v = '[![](.%(fn)s.svg)](%(loc)scall_sequence.html?src=%(fn)s)' % m
    #     v = '[![](.%(fn)s.svg)](?src=%(fn)s&sequence_details=true)' % m
    #     v = T.closed_admon('Call Sequence', v, 'note')
    # elif viz == Mermaid:
    #     v = '{!%(fn)s.plantuml!}' % m
    # add_doc('\n%s\n' % v)
